        #wv_index_list = [self.input_psfs_wvs.index(wv) for wv in wvs]
__author__ = 'jruffio'
import multiprocessing as mp
import ctypes

import numpy as np
import pyklip.spectra_management as spec
import os
import itertools

from pyklip.fmlib.nofm import NoFM
import pyklip.fm as fm

from scipy import interpolate
from copy import copy

import astropy.io.fits as pyfits

from time import time
import scipy.ndimage as ndimage

debug = False

class MatchedFilter(NoFM):
    """
    Matched filter with forward modelling.
    """
    def __init__(self, inputs_shape,
                 numbasis,
                 input_psfs,
                 input_psfs_wvs,
                 spectrallib,
                 save_per_sector = None,
                 datatype="float",
                 fakes_sepPa_list = None,
                 disable_FM = None,
                 true_fakes_pos = None,
                 ref_center = None,
                 flipx = None,
                 rm_edge = None,
                 edge_threshold = None,
                 planet_radius = None,
                 background_width = None,
                 save_bbfm = None,
                 save_fm = None,
                 save_fmout_path = None):
        '''
        Defining the forward model matched filter parameters

        Args:
            inputs_shape: shape of the inputs numpy array. Typically (N, y, x)
            numbasis: 1d numpy array consisting of the number of basis vectors to use
            input_psfs: the psf of the image. A numpy array with shape (wv, y, x)
            input_psfs_wvs: the wavelegnths that correspond to the input psfs
            spectrallib: if not None, a list of spectra in raw DN units. The spectra should:
                    - have the total flux of the star, ie correspond to a contrast of 1.
                    - represent the total flux of the PSF and not the simply peak value.
                    - be multiplied by the atmospheric and instrumental transmission.
                    - have the same size as the number of images in the dataset.
            save_per_sector: If not None, should be a filename where the fmout array will be saved after each sector.
                    (Caution: huge file!! easily tens of Gb.)
            datatype: datatype to be used for the numpy arrays: "double" or "float" (default).
            fakes_sepPa_list: [(sep_pix1,pa1),(sep_pix2,pa2),...].
                    List of separations and pas for the simulated planets in the data.
                    If not None, it will only calculated the matched filter at the position of the fakes and skip the rest.
            disable_FM: Disable the calculation of the forward model in the code.
                        The unchanged original PSF will be used instead. (Default False)
            true_fakes_pos: If True and fakes_only is True, calculate the forward model at the exact position of the
                    fakes and not at the center of the pixels. (Default False)
            ref_center: reference center to which all the images are aligned. It should be the same center as the one
                    used in fm.parallelized.
            flipx: Determines whether a relfection about the x axis is necessary to rotate image North-up East left.
                    Should match the same attribute in the instrument class.
            rm_edge: When True (default), remove image edges to avoid edge effect. When there is more than 25% of NaNs
                    in the projection of the FM model on the data, the result of the projection is set to NaNs right away.
            edge_threshold: if rm_edge is true, defines the fraction of pixels that needs to be finite in the aperture
                around the planet for the projection to be calculated.
            planet_radius: Radius of the aperture to be used for the matched filter (pick something of the order of the
                            2xFWHM)
            background_width: Half the width of the arc in which the local standard deviation will be calculated.
            save_bbfm: If true, saves the broadband forward models
            savebfm: If true, saves all the forward models. This is only tractable for extremely small dataset.
            save_fmout_path: If not None, path to a directory where to save fmout at the end.

        '''
        # allocate super class
        super(MatchedFilter, self).__init__(inputs_shape, np.array(numbasis))

        self.save_bbfm = save_bbfm
        self.save_fm = save_fm
        if (self.save_bbfm or self.save_fm) and (np.size(numbasis) != 1 or len(spectrallib) != 1):
            raise ValueError("save_bbfm or save_fm only work when np.size(numbasis)==1 and len(self.spectrallib)==1 ")
        if (self.save_bbfm and self.save_fm):
            raise ValueError("save_bbfm or save_fm cannot be used at the same time ")
        self.save_fmout_path = save_fmout_path

        if rm_edge is not None:
            self.rm_edge = rm_edge
        else:
            self.rm_edge = True

        if edge_threshold is None:
            self.edge_threshold = 0.75
        else:
            self.edge_threshold = edge_threshold

        if true_fakes_pos is None:
            self.true_fakes_pos = False
        else:
            self.true_fakes_pos = true_fakes_pos

        if datatype=="double":
            self.data_type = ctypes.c_double
        elif datatype=="float":
            self.data_type = ctypes.c_float

        if save_per_sector is not None:
            self.fmout_dir = save_per_sector
            self.save_raw_fmout = True
        else:
            self.save_raw_fmout = False

        self.N_numbasis =  np.size(numbasis)
        self.ny = self.inputs_shape[1]
        self.nx = self.inputs_shape[2]
        self.N_frames = self.inputs_shape[0]

        self.fakes_sepPa_list = fakes_sepPa_list
        if disable_FM is None:
            self.disable_FM = False
        else:
            self.disable_FM = disable_FM

        self.inputs_shape = self.inputs_shape

        self.input_psfs_wvs = list(np.array(input_psfs_wvs,dtype=self.data_type))

        # Make sure the total flux of each PSF is unity for all wavelengths
        # So the peak value won't be unity.
        self.input_psfs = input_psfs/np.nansum(input_psfs,axis=(1,2))[:,None,None]
        numwv_psf,ny_psf,nx_psf =  self.input_psfs.shape

        self.spectrallib = spectrallib
        self.N_spectra = len(self.spectrallib)


        # create bounds for PSF stamp size
        self.row_m = int(np.floor(ny_psf/2.0))    # row_minus
        self.row_p = int(np.ceil(ny_psf/2.0))     # row_plus
        self.col_m = int(np.floor(nx_psf/2.0))    # col_minus
        self.col_p = int(np.ceil(nx_psf/2.0))     # col_plus

        self.psf_centx_notscaled = {}
        self.psf_centy_notscaled = {}
        self.curr_pa_fk = {}
        self.curr_sep_fk = {}

        x_psf_grid, y_psf_grid = np.meshgrid(np.arange(nx_psf * 1.)-nx_psf//2,np.arange(ny_psf* 1.)-ny_psf//2)
        psfs_func_list = []
        self.input_psfs[np.where(np.isnan(self.input_psfs))] = 0
        import warnings
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            for wv_index in range(numwv_psf):
                model_psf = self.input_psfs[wv_index, :, :]
                psf_func = interpolate.LSQBivariateSpline(x_psf_grid.ravel(),y_psf_grid.ravel(),model_psf.ravel(),x_psf_grid[0,0:nx_psf-1]+0.5,y_psf_grid[0:ny_psf-1,0]+0.5)
                psfs_func_list.append(psf_func)

        self.psfs_func_list = psfs_func_list

        ny_PSF,nx_PSF = input_psfs.shape[1:]
        if ny_PSF < 8 or nx_PSF < 8:
            raise Exception("PSF cube is too small. It needs a stamp width bigger than 8 pixels.")
        stamp_PSF_x_grid, stamp_PSF_y_grid = np.meshgrid(np.arange(0,nx_PSF,1)-nx_PSF//2,np.arange(0,ny_PSF,1)-ny_PSF//2)
        self.stamp_PSF_mask = np.ones((ny_PSF,nx_PSF))
        r_PSF_stamp = np.sqrt((stamp_PSF_x_grid)**2 +(stamp_PSF_y_grid)**2)
        if planet_radius is not None:
            self.planet_radius = planet_radius
        else:
            self.planet_radius =  int(np.round(np.min([ny_PSF,nx_PSF])*7./20.))
        self.stamp_PSF_mask[np.where(r_PSF_stamp < self.planet_radius)] = np.nan
        # self.stamp_PSF_mask[np.where(r_PSF_stamp < 4.)] = np.nan
        if background_width is not None:
            self.background_width = background_width
        else:
            self.background_width =  np.min([ny_PSF,nx_PSF])//2
        self.bbfm_mask = np.ones((self.planet_radius*2,self.planet_radius*2))
        stamp_bbfm_x_grid, stamp_bbfm_y_grid = np.meshgrid(np.arange(0,self.planet_radius*2,1)-self.planet_radius,
                                                           np.arange(0,self.planet_radius*2,1)-self.planet_radius)
        r_bbfm_stamp = abs((stamp_bbfm_x_grid) +(stamp_bbfm_y_grid)*1j)
        self.bbfm_mask[np.where(r_bbfm_stamp < self.planet_radius)] = np.nan

        if ref_center is not None:
            self.ref_center = ref_center
            # create some parameters for a blank canvas to draw psfs on
            nx = inputs_shape[2]
            ny = inputs_shape[1]
            self.x_grid, self.y_grid = np.meshgrid(np.arange(nx * 1.)-ref_center[0], np.arange(ny * 1.)-ref_center[1])
            self.r_grid = abs(self.x_grid +self.y_grid*1j)
            self.pa_grid = np.arctan2( -self.x_grid,self.y_grid) % (2.0 * np.pi)
            if flipx is not None:
                sign = -1.
                if flipx:
                    sign = 1.
                self.th0_grid = np.arctan2(sign*self.x_grid,self.y_grid)


    # def alloc_interm(self, max_sector_size, numsciframes):
    #     """Allocates shared memory array for intermediate step
    #
    #     Intermediate step is allocated for a sector by sector basis
    #
    #     Args:
    #         max_sector_size: number of pixels in this sector. Max because this can be variable. Stupid rotating sectors
    #
    #     Returns:
    #         interm: mp.array to store intermediate products from one sector in
    #         interm_shape:shape of interm array (used to convert to numpy arrays)
    #
    #     """
    #     if self.save_bbfm is not None:
    #         interm_size = self.planet_radius*self.planet_radius*self.ny*self.nx
    #         interm = mp.Array(ctypes.c_double, interm_size)
    #         interm_shape = [self.ny,self.nx,self.planet_radius,self.planet_radius]
    #
    #         return interm, interm_shape
    #     else:
    #         return None,None



    def alloc_fmout(self, output_img_shape):
        """
        Allocates shared memory for the output of the shared memory

        Args:
            output_img_shape: Not used

        Returns:
            fmout: mp.array to store auxilliary data in
            fmout_shape: shape of auxilliary array = (3,self.N_spectra,self.N_numbasis,self.N_frames,self.ny,self.nx)
                        The 3 is for saving the different term of the matched filter:
                            0: dot product
                            1: square of the norm of the model
                            2: Local estimated variance of the data
                            3: Number of pixels used in the matched filter

        """
        if self.save_bbfm:
            fmout_size = 4*self.N_spectra*self.N_numbasis*self.N_frames*self.ny*self.nx + \
                         2*self.planet_radius*2*self.planet_radius*self.ny*self.nx
            fmout = mp.Array(self.data_type, fmout_size)
            fmout_shape = (4*self.N_spectra*self.N_numbasis*self.N_frames*self.ny*self.nx + \
                         2*self.planet_radius*2*self.planet_radius*self.ny*self.nx,1)
        elif self.save_fm:
            fmout_size = 4*self.N_spectra*self.N_numbasis*self.N_frames*self.ny*self.nx + \
                         self.N_frames*2*self.planet_radius*2*self.planet_radius*self.ny*self.nx
            fmout = mp.Array(self.data_type, fmout_size)
            fmout_shape = (4*self.N_spectra*self.N_numbasis*self.N_frames*self.ny*self.nx + \
                         self.N_frames*2*self.planet_radius*2*self.planet_radius*self.ny*self.nx,1)
        else:
            fmout_size = 4*self.N_spectra*self.N_numbasis*self.N_frames*self.ny*self.nx
            fmout = mp.Array(self.data_type, fmout_size)
            fmout_shape = (4,self.N_spectra,self.N_numbasis,self.N_frames,self.ny,self.nx)


        return fmout, fmout_shape

    def skip_section(self, radstart, radend, phistart, phiend,flipx=True):
        """
        Returns a boolean indicating if the section defined by (radstart, radend, phistart, phiend) should be skipped.
        When True is returned the current section in the loop in klip_parallelized() is skipped.

        Args:
            radstart: minimum radial distance of sector [pixels]
            radend: maximum radial distance of sector [pixels]
            phistart: minimum azimuthal coordinate of sector [radians]
            phiend: maximum azimuthal coordinate of sector [radians]
            flipx: if True, flip x coordinate in final image

        Returns:
            Boolean: False so by default it never skips.
        """

        margin_sep = np.sqrt(2)/2.
        margin_phi = np.sqrt(2)/(2*radstart)
        if self.fakes_sepPa_list is not None:
            skipSectionBool = True
            for sep_it,pa_it in self.fakes_sepPa_list:
                if flipx:
                    paend= ((-phistart + np.pi/2.)% (2.0 * np.pi))
                    pastart = ((-phiend + np.pi/2.)% (2.0 * np.pi))
                else:
                    pastart = ((phistart - np.pi/2.)% (2.0 * np.pi))
                    paend= ((phiend - np.pi/2.)% (2.0 * np.pi))
                # Normal case when there are no 2pi wrap
                if pastart < paend:
                    if (radstart-margin_sep<=sep_it<=radend+margin_sep) and ((pa_it%360)/180.*np.pi >= pastart-margin_phi) & ((pa_it%360)/180.*np.pi < paend+margin_phi):
                        skipSectionBool = False
                        break
                # 2 pi wrap case
                else:
                    if (radstart-margin_sep<=sep_it<=radend+margin_sep) and (((pa_it%360)/180.*np.pi >= pastart-margin_phi) | ((pa_it%360)/180.*np.pi < paend+margin_phi)):
                        skipSectionBool = False
                        break
        else:
            skipSectionBool = False

        return skipSectionBool


    def fm_from_eigen(self, klmodes=None, evals=None, evecs=None, input_img_shape=None, input_img_num=None,
                      ref_psfs_indicies=None, section_ind=None,section_ind_nopadding=None, aligned_imgs=None, pas=None,
                     wvs=None, radstart=None, radend=None, phistart=None, phiend=None, padding=None,IOWA = None, ref_center=None,
                     parang=None, ref_wv=None, numbasis=None, fmout=None, perturbmag=None,klipped=None, flipx=True, **kwargs):
        """
        Calculate and project the FM at every pixel of the sector. Store the result in fmout.

        Args:
            klmodes: unpertrubed KL modes
            evals: eigenvalues of the covariance matrix that generated the KL modes in ascending order
                   (lambda_0 is the 0 index) (shape of [nummaxKL])
            evecs: corresponding eigenvectors (shape of [p, nummaxKL])
            input_img_shape: 2-D shape of inpt images ([ysize, xsize])
            input_img_num: index of sciece frame
            ref_psfs_indicies: array of indicies for each reference PSF
            section_ind: array indicies into the 2-D x-y image that correspond to this section.
                         Note needs be called as section_ind[0]
            pas: array of N parallactic angles corresponding to N reference images [degrees]
            wvs: array of N wavelengths of those referebce images
            radstart: radius of start of segment
            radend: radius of end of segment
            phistart: azimuthal start of segment [radians]
            phiend: azimuthal end of segment [radians]
            padding: amount of padding on each side of sector
            IOWA: tuple (IWA,OWA) where IWA = Inner working angle and OWA = Outer working angle both in pixels.
                It defines the separation interva in which klip will be run.
            ref_center: center of image
            numbasis: array of KL basis cutoffs
            parang: parallactic angle of input image [DEGREES]
            ref_wv: wavelength of science image
            fmout: numpy output array for FM output. Shape is (N, y, x, b)
            klipped: array of shape (p,b) that is the PSF subtracted data for each of the b KLIP basis
                     cutoffs. If numbasis was an int, then sub_img_row_selected is just an array of length p
            kwargs: any other variables that we don't use but are part of the input
        """
        if hasattr(self,"ref_center"):
            if (self.ref_center[0] != ref_center[0]) or (self.ref_center[1] != ref_center[1]):
                raise ValueError("ref_center needs to be the same when defining the matchedFilter class and calling parallelized")
        if np.size(numbasis) != 1:
            raise ValueError("Numbasis should only have a single element. e.g. numbasis = [30]. numbasis = [10,20,30] is not accepted.")

        if self.save_bbfm:
            fmout1 = fmout[:4*self.N_spectra*self.N_numbasis*self.N_frames*self.ny*self.nx]
            fmout1.shape = (4,self.N_spectra,self.N_numbasis,self.N_frames,self.ny,self.nx)
            fmout2 = fmout[4*self.N_spectra*self.N_numbasis*self.N_frames*self.ny*self.nx::]
            fmout2.shape = (self.ny,self.nx,2*self.planet_radius,2*self.planet_radius)
        elif self.save_fm:
            fmout1 = fmout[:4*self.N_spectra*self.N_numbasis*self.N_frames*self.ny*self.nx]
            fmout1.shape = (4,self.N_spectra,self.N_numbasis,self.N_frames,self.ny,self.nx)
            fmout2 = fmout[4*self.N_spectra*self.N_numbasis*self.N_frames*self.ny*self.nx::]
            fmout2.shape = (self.N_frames,self.ny,self.nx,2*self.planet_radius,2*self.planet_radius)
        else:
            fmout1 = fmout

        ref_wv = ref_wv.astype(self.data_type)

        sci = aligned_imgs[input_img_num, section_ind[0]]
        refs = aligned_imgs[ref_psfs_indicies, :]
        refs = refs[:, section_ind[0]]

        # Calculate the PA,sep 2D map

        if hasattr(self,"x_grid") and hasattr(self,"y_grid"):
            x_grid, y_grid = self.x_grid,self.y_grid
        else:
            x_grid, y_grid = np.meshgrid(np.arange(self.nx * 1.)-ref_center[0], np.arange(self.ny * 1.)-ref_center[1])
        x_grid=x_grid.astype(self.data_type)
        y_grid=y_grid.astype(self.data_type)
        # Define the masks for where the planet is and the background.
        if hasattr(self,"r_grid"):
            r_grid = self.r_grid
        else:
            r_grid = np.sqrt((x_grid)**2 + (y_grid)**2)
        if hasattr(self,"pa_grid"):
            pa_grid = self.pa_grid
        else:
            pa_grid = np.arctan2( -x_grid,y_grid) % (2.0 * np.pi)
        if flipx:
            paend= ((-phistart + np.pi/2.)% (2.0 * np.pi))
            pastart = ((-phiend + np.pi/2.)% (2.0 * np.pi))
        else:
            pastart = ((phistart - np.pi/2.)% (2.0 * np.pi))
            paend= ((phiend - np.pi/2.)% (2.0 * np.pi))
        # Normal case when there are no 2pi wrap
        if pastart < paend:
            where_section = np.where((r_grid >= radstart) & (r_grid < radend) & (pa_grid >= pastart) & (pa_grid < paend))
        # 2 pi wrap case
        else:
            where_section = np.where((r_grid >= radstart) & (r_grid < radend) & ((pa_grid >= pastart) | (pa_grid < paend)))

        # Get a list of the PAs and sep of the PA,sep map falling in the current section
        r_list = r_grid[where_section]
        pa_list = pa_grid[where_section]
        x_list = x_grid[where_section]
        y_list = y_grid[where_section]
        row_id_list = where_section[0]
        col_id_list = where_section[1]
        # Only select pixel with fakes if needed
        if self.fakes_sepPa_list is not None:
            r_list_tmp = []
            pa_list_tmp = []
            row_id_list_tmp = []
            col_id_list_tmp = []
            for sep_it,pa_it in self.fakes_sepPa_list:
                x_it = sep_it*np.cos(np.radians(90+pa_it))
                y_it = sep_it*np.sin(np.radians(90+pa_it))
                dist_list = np.sqrt((x_list-x_it)**2+(y_list-y_it)**2)
                min_id = np.nanargmin(dist_list)
                min_dist = dist_list[min_id]
                if min_dist < np.sqrt(2)/2.:
                    if self.true_fakes_pos:
                        r_list_tmp.append(sep_it)
                        pa_list_tmp.append(np.radians(pa_it))
                    else:
                        r_list_tmp.append(r_list[min_id])
                        pa_list_tmp.append(pa_list[min_id])
                    row_id_list_tmp.append(row_id_list[min_id])
                    col_id_list_tmp.append(col_id_list[min_id])
            r_list = r_list_tmp
            pa_list = pa_list_tmp
            row_id_list = row_id_list_tmp
            col_id_list = col_id_list_tmp

        greenboard = np.zeros((self.ny,self.nx))
        x_bbfm, y_bbfm = np.meshgrid(np.arange(self.ny, dtype=np.float32), np.arange(self.nx, dtype=np.float32))
        # flip x if needed to get East left of North
        if flipx is True:
            x_bbfm = ref_center[0] - (x_bbfm - ref_center[0])
        # do rotation. CW rotation formula to get a CCW of the image
        angle_rad = np.radians(parang)
        cosa = np.cos(angle_rad)
        sina = np.sin(angle_rad)
        xp = (x_bbfm-ref_center[0])*cosa + (y_bbfm-ref_center[1])*sina + ref_center[0]
        yp = -(x_bbfm-ref_center[0])*sina + (y_bbfm-ref_center[1])*cosa + ref_center[1]
        # create bounds for PSF stamp size
        self.bbfm_m = int(np.floor(self.planet_radius))
        self.bbfm_p = int(np.ceil(self.planet_radius))

        # Loop over the input template spectra and the number of KL modes in numbasis
        for spec_id,N_KL_id in itertools.product(range(self.N_spectra),range(self.N_numbasis)):
            # Calculate the projection of the FM and the klipped section for every pixel in the section.
            # 1/ Inject a fake at one pa and sep in the science image
            # 2/ Inject the corresponding planets at the same PA and sep in the reference images remembering that the
            # references rotate.
            # 3/ Calculate the perturbation of the KL modes
            # 4/ Calculate the FM
            # 5/ Calculate dot product (matched filter)
            for sep_fk,pa_fk,row_id,col_id in zip(r_list,np.rad2deg(pa_list),row_id_list,col_id_list):
                # print(sep_fk,pa_fk,row_id,col_id)
                # 1/ Inject a fake at one pa and sep in the science image
                try:
                    model_sci,mask = self.generate_model_sci(input_img_shape, section_ind, parang, ref_wv,
                                                             radstart, radend, phistart, phiend, padding, ref_center,
                                                             parang, ref_wv,sep_fk,pa_fk, flipx)
                except:
                    # print("skip",sep_fk,pa_fk)
                    continue
                # Normalize the science image according to the spectrum. the model is normalize to unit contrast,
                model_sci = model_sci*self.spectrallib[spec_id][input_img_num]
                where_fk = np.where(mask==2)[0]
                where_background = np.where(mask>=1)[0] # Caution: it includes where the fake is...
                where_background_strict = np.where(mask==1)[0]

                # JB hack trying to include covariances
                # if 0:
                #     section_ind_y,section_ind_x = np.unravel_index(section_ind,(self.ny,self.nx))
                #     section_ind_y,section_ind_x = section_ind_y[0][where_fk],section_ind_x[0][where_fk]
                #     distances = np.sqrt((section_ind_x[:,None]-section_ind_x[None,:])**2+(section_ind_y[:,None]-section_ind_y[None,:])**2)
                #     from  scipy.interpolate import interp1d
                #     corr_profile = interp1d([0,1,np.sqrt(2),2],[1,0.2,-0.1,0],bounds_error=False, fill_value=0)
                #     spatial_covar =corr_profile(distances)
                #     inv_spatial_covar = np.linalg.inv(spatial_covar)
                #
                #     # import matplotlib.pyplot as plt
                #     # plt.subplot(1,3,1)
                #     # plt.imshow(distances,interpolation="nearest")
                #     # plt.subplot(1,3,2)
                #     # plt.imshow(spatial_covar,interpolation="nearest")
                #     # plt.subplot(1,3,3)
                #     # plt.imshow(np.linalg.inv(spatial_covar),interpolation="nearest")
                #     # plt.show()
                #     # exit()

                if np.size(klipped[where_fk,N_KL_id]) == 0:
                    quitnow = True
                elif self.rm_edge and float(np.sum(np.isfinite(klipped[where_fk,N_KL_id])))/float(np.size(klipped[where_fk,N_KL_id]))<=self.edge_threshold:
                    quitnow = True
                else:
                    quitnow = False

                if quitnow:
                    fmout1[0,spec_id,N_KL_id,input_img_num,row_id,col_id] = np.nan
                    fmout1[1,spec_id,N_KL_id,input_img_num,row_id,col_id] = np.nan
                    fmout1[2,spec_id,N_KL_id,input_img_num,row_id,col_id] = np.nan
                    fmout1[3,spec_id,N_KL_id,input_img_num,row_id,col_id] = np.nan
                    continue

                # 2/ Inject the corresponding planets at the same PA and sep in the reference images remembering that the
                # references rotate.
                if not self.disable_FM:
                    models_ref = self.generate_models(input_img_shape, section_ind, pas, wvs, radstart, radend,
                                                      phistart, phiend, padding, ref_center, parang, ref_wv,sep_fk,pa_fk, flipx)

                    # Normalize the models with the spectrum. the model is normalize to unit contrast,
                    input_spectrum = self.spectrallib[spec_id][ref_psfs_indicies]
                    models_ref = models_ref * input_spectrum[:, None]

                    # 3/ Calculate the perturbation of the KL modes
                    # using original Kl modes and reference models, compute the perturbed KL modes.
                    # Spectrum is already in the model, that's why we use perturb_specIncluded(). (Much faster)
                    delta_KL = fm.perturb_specIncluded(evals, evecs, klmodes, refs, models_ref)

                    # 4/ Calculate the FM: calculate postklip_psf using delta_KL
                    # postklip_psf has unit broadband contrast
                    model_sci_fk = model_sci[where_fk]
                    delta_KL_kl = delta_KL[:,where_fk]
                    klmodes_fk = klmodes[:,where_fk]
                    postklip_psf = calculate_fm_opti(delta_KL, klmodes,sci, model_sci_fk,delta_KL_kl,klmodes_fk)
                    # postklip_psf, oversubtraction, selfsubtraction = fm.calculate_fm(delta_KL, klmodes, numbasis,
                    #                                                                  sci, model_sci, inputflux=None)
                else:
                    #if one doesn't want the FM
                    if np.size(numbasis) == 1:
                        postklip_psf = model_sci[None,:]
                    else:
                        # Mh, is this actually working? shouldn't I use tile?
                        postklip_psf = model_sci

                # 5/ Calculate dot product (matched filter)
                # fmout_shape = (3,self.N_spectra,self.N_numbasis,self.N_frames,self.ny,self.nx)
                #         The 3 is for saving the different term of the matched filter:
                #             0: dot product
                #             1: square of the norm of the model
                #             2: Local estimated variance of the data
                sky = np.nanmean(klipped[where_background_strict,N_KL_id])
                # postklip_psf[N_KL_id,where_fk] = postklip_psf[N_KL_id,where_fk]-np.mean(postklip_psf[N_KL_id,where_background])
                # Subtract local sky background to the klipped image
                klipped_sub = klipped[where_fk,N_KL_id]-sky
                # klipped_sub_finite = np.where(np.isfinite(klipped_sub))
                # klipped_sub_nan = np.where(np.isnan(klipped_sub))
                postklip_psf[N_KL_id,np.where(np.isnan(klipped_sub))[0]] = np.nan
                postklip_psf_fk = postklip_psf[N_KL_id,:]
                dot_prod = np.nansum(klipped_sub*postklip_psf_fk)
                model_norm = np.nansum(postklip_psf_fk*postklip_psf_fk)
                klipped_rm_pl = copy(klipped[:,N_KL_id]) -sky
                klipped_rm_pl[where_fk] -=  (dot_prod/model_norm)*postklip_psf_fk
                klipped_rm_pl_bkg = klipped_rm_pl[where_background]
                if self.rm_edge and (float(np.sum(np.isfinite(klipped_rm_pl_bkg)))/float(np.size(klipped_rm_pl_bkg))<=self.edge_threshold):
                    variance = np.nan
                    npix = np.nan
                else:
                    variance = np.nanvar(klipped_rm_pl_bkg)
                    npix = np.sum(np.isfinite(klipped_rm_pl_bkg))

                # JB hack trying to include covariances
                # else:
                #     dot_prod = np.nansum(klipped_sub*postklip_psf_fk)
                #     model_norm = np.nansum(postklip_psf_fk*postklip_psf_fk)
                #     klipped_rm_pl = copy(klipped[:,N_KL_id]) -sky
                #     klipped_rm_pl[where_fk] -=  (dot_prod/model_norm)*postklip_psf_fk
                #     klipped_rm_pl_bkg = klipped_rm_pl[where_background]
                #     if self.rm_edge and (float(np.sum(np.isfinite(klipped_rm_pl_bkg)))/float(np.size(klipped_rm_pl_bkg))<=self.edge_threshold):
                #         variance = np.nan
                #         npix = np.nan
                #     else:
                #         variance = np.nanvar(klipped_rm_pl_bkg)
                #         npix = np.sum(np.isfinite(klipped_rm_pl_bkg))
                #
                #     inv_spatial_covar = inv_spatial_covar/variance
                #     # print(klipped_sub.shape,inv_spatial_covar.shape)
                #     # print(np.dot(klipped_sub,inv_spatial_covar).shape)
                #     dot_prod = np.nansum(np.dot(klipped_sub,inv_spatial_covar)*postklip_psf_fk)
                #     model_norm = np.nansum(np.dot(postklip_psf_fk,inv_spatial_covar)*postklip_psf_fk)
                #     variance = 1.0
                #     # exit()



                fmout1[0,spec_id,N_KL_id,input_img_num,row_id,col_id] = dot_prod
                fmout1[1,spec_id,N_KL_id,input_img_num,row_id,col_id] = model_norm
                fmout1[2,spec_id,N_KL_id,input_img_num,row_id,col_id] = variance
                fmout1[3,spec_id,N_KL_id,input_img_num,row_id,col_id] = npix

                if self.save_bbfm:
                    greenboard.shape = [input_img_shape[0] * input_img_shape[1]]
                    greenboard[section_ind[0][where_fk]] = postklip_psf[N_KL_id,:]
                    greenboard.shape = [input_img_shape[0],input_img_shape[1]]

                    rot_stamp = ndimage.map_coordinates(greenboard,
                                                        [yp[(row_id-self.bbfm_m):(row_id+self.bbfm_p),
                                                            (col_id-self.bbfm_m):(col_id+self.bbfm_p)].ravel(),
                                                         xp[(row_id-self.bbfm_m):(row_id+self.bbfm_p),
                                                            (col_id-self.bbfm_m):(col_id+self.bbfm_p)].ravel()],
                                                        cval=np.nan)
                    rot_stamp.shape = [2*self.planet_radius,2*self.planet_radius]

                    fmout2[row_id,col_id,:,:] = fmout2[row_id,col_id,:,:]+rot_stamp
                elif self.save_fm:
                    greenboard.shape = [input_img_shape[0] * input_img_shape[1]]
                    greenboard[section_ind[0][where_fk]] = postklip_psf[N_KL_id,:]
                    greenboard.shape = [input_img_shape[0],input_img_shape[1]]
                    greenboard_pad = np.pad(greenboard,((self.bbfm_m,self.bbfm_p),(self.bbfm_m,self.bbfm_p)),mode="constant",constant_values=np.nan)
                    mystamp = copy(greenboard_pad[(row_id):(row_id+self.bbfm_m+self.bbfm_p),
                                                            (col_id):(col_id+self.bbfm_m+self.bbfm_p)])
                    # print((row_id-self.bbfm_m),(row_id+self.bbfm_p),
                    #                                         (col_id-self.bbfm_m),(col_id+self.bbfm_p))
                    # import matplotlib.pyplot as plt
                    # plt.subplot(1,2,1)
                    # plt.imshow(greenboard, interpolation="nearest")
                    # plt.subplot(1,2,2)
                    # plt.imshow(mystamp, interpolation="nearest")
                    # plt.show()

                    fmout2[input_img_num,row_id,col_id,:,:] = mystamp

                if 0:
                    print(dot_prod,model_norm,variance,npix)

                # Plot sector, klipped and FM model for debug only
                if 0: # and row_id>=10 and col_id > 5 # and np.nansum(klipped[where_fk,N_KL_id]) != 0:
                    # plt.figure(1)
                    # blackboard = np.zeros((self.ny,self.nx))
                    # blackboard.shape = [input_img_shape[0] * input_img_shape[1]]
                    # blackboard[section_ind] = klipped
                    # blackboard.shape = [input_img_shape[0],input_img_shape[1]]
                    # plt.imshow(blackboard,interpolation="nearest")
                    # plt.colorbar()
                    # plt.figure(2)
                    # for k in range(numbasis[0]):
                    #     blackboard = np.zeros((self.ny,self.nx))
                    #     blackboard.shape = [input_img_shape[0] * input_img_shape[1]]
                    #     blackboard[section_ind] = klmodes[k,:]
                    #     blackboard.shape = [input_img_shape[0],input_img_shape[1]]
                    #     plt.subplot(1,numbasis[0],k+1)
                    #     plt.imshow(blackboard[::-1,:],interpolation="nearest")
                    #     plt.title("KL{0}".format(k))
                    #     plt.colorbar()
                    # plt.show()


                    #if 0:
                    # print(klipped_sub)
                    # print(np.isfinite(klipped_sub))
                    # print(np.size(klipped_sub))
                    # print(float(np.sum(np.isfinite(klipped_sub)))/float(np.size(klipped_sub)))
                    # print(float(np.sum(np.isfinite(klipped[where_background,N_KL_id])))/float(np.size(klipped[where_background,N_KL_id])))
                    print(sep_fk,pa_fk,row_id,col_id)
                    print(dot_prod,model_norm,variance,npix)
                    # print(np.nanmean(klipped-sky),sky,dot_prod,model_norm,np.nanmean((dot_prod/model_norm)*postklip_psf[N_KL_id,:]))
                    # print(klipped.shape,postklip_psf[N_KL_id,:].shape)
                    # print(float(np.sum(np.isfinite(klipped_rm_pl[where_background]))),float(np.size(klipped_rm_pl[where_background])))
                    blackboard1 = np.zeros((self.ny,self.nx))
                    blackboard2 = np.zeros((self.ny,self.nx))
                    blackboard3 = np.zeros((self.ny,self.nx))
                    #print(section_ind)
                    plt.figure(1)
                    plt.subplot(1,3,1)
                    blackboard1.shape = [input_img_shape[0] * input_img_shape[1]]
                    blackboard1[section_ind] = mask
                    blackboard1[section_ind] = blackboard1[section_ind] + 1
                    blackboard1.shape = [input_img_shape[0],input_img_shape[1]]
                    plt.imshow(blackboard1,interpolation="nearest")
                    plt.colorbar()
                    plt.subplot(1,3,2)
                    blackboard2.shape = [input_img_shape[0] * input_img_shape[1]]
                    # blackboard2[section_ind[0][where_fk]] = klipped[where_fk,N_KL_id]
                    blackboard2[section_ind[0]] = klipped#klipped_rm_pl
                    blackboard2.shape = [input_img_shape[0],input_img_shape[1]]
                    plt.imshow(blackboard2,interpolation="nearest")
                    plt.colorbar()
                    plt.subplot(1,3,3)
                    blackboard3.shape = [input_img_shape[0] * input_img_shape[1]]
                    blackboard3[section_ind[0][where_fk]] = postklip_psf[N_KL_id,:]
                    blackboard3.shape = [input_img_shape[0],input_img_shape[1]]
                    plt.imshow(blackboard3,interpolation="nearest")
                    plt.colorbar()
                    #print(klipped[where_fk,N_KL_id])
                    #print(postklip_psf[N_KL_id,where_fk])
                    # print(np.sum(klipped[where_fk,N_KL_id]*postklip_psf[N_KL_id,where_fk]))
                    # print(np.sum(postklip_psf[N_KL_id,where_fk]*postklip_psf[N_KL_id,where_fk]))
                    # print(np.sum(klipped[where_fk,N_KL_id]*klipped[where_fk,N_KL_id]))
                    plt.show()
                # exit()


    def fm_end_sector(self, interm_data=None, fmout=None, sector_index=None,
                               section_indicies=None):
        """
        Save the fmout object at the end of each sector if save_per_sector was defined when initializing the class.
        """
        #fmout_shape = (3,self.N_spectra,self.N_numbasis,self.N_frames,self.ny,self.nx)
        if self.save_raw_fmout:
            if self.save_bbfm or self.save_fm:
                fmout1 = fmout[:4*self.N_spectra*self.N_numbasis*self.N_frames*self.ny*self.nx]
                fmout1.shape = (4,self.N_spectra,self.N_numbasis,self.N_frames,self.ny,self.nx)
            else:
                fmout1 = fmout
            hdu = pyfits.PrimaryHDU(fmout)
            hdulist = pyfits.HDUList([hdu])
            hdulist.writeto(self.fmout_dir,clobber=True)


        return

    def save_fmout(self, dataset, fmout, outputdir, fileprefix, numbasis, klipparams=None, calibrate_flux=False,
                   spectrum=None, pixel_weights=1):
        """
        Saves the fmout data to disk following the instrument's savedata function

        Args:
            dataset: Instruments.Data instance. Will use its dataset.savedata() function to save data
            fmout: the fmout data passed from fm.klip_parallelized which is passed as the output of cleanup_fmout
            outputdir: output directory
            fileprefix: the fileprefix to prepend the file name
            numbasis: KL mode cutoffs used
            klipparams: string with KLIP-FM parameters
            calibrate_flux: if True, flux calibrate the data (if applicable)
            spectrum: if not None, the spectrum to weight the data by. Length same as dataset.wvs
            pixel_weights: weights for each pixel for weighted mean. [Not used in matched filter]
        """

        if self.save_bbfm:
            fmout1 = fmout[:4*self.N_spectra*self.N_numbasis*self.N_frames*self.ny*self.nx]
            fmout1.shape = (4,self.N_spectra,self.N_numbasis,self.N_frames,self.ny,self.nx)
            fmout2 = fmout[4*self.N_spectra*self.N_numbasis*self.N_frames*self.ny*self.nx::]
            fmout2.shape = (self.ny,self.nx,2*self.planet_radius,2*self.planet_radius)
        elif self.save_fm:
            fmout1 = fmout[:4*self.N_spectra*self.N_numbasis*self.N_frames*self.ny*self.nx]
            fmout1.shape = (4,self.N_spectra,self.N_numbasis,self.N_frames,self.ny,self.nx)
            fmout2 = fmout[4*self.N_spectra*self.N_numbasis*self.N_frames*self.ny*self.nx::]
            fmout2.shape = (self.N_frames,self.ny,self.nx,2*self.planet_radius,2*self.planet_radius)
        else:
            fmout1 = fmout

        if self.save_fmout_path is not None:
            hdu = pyfits.PrimaryHDU(fmout1)
            hdulist = pyfits.HDUList([hdu])
            suffix = "fmout"
            try:
                hdulist.writeto(os.path.join(self.save_fmout_path,fileprefix+'-'+suffix+'.fits'), overwrite=True)
            except TypeError:
                hdulist.writeto(os.path.join(self.save_fmout_path,fileprefix+'-'+suffix+'.fits'), clobber=True)

        #fmout_shape = (3,self.N_spectra,self.N_numbasis,self.N_frames,self.ny,self.nx)
        fmout1[np.where(fmout1==0)] = np.nan

        # hdu = pyfits.PrimaryHDU(fmout1)
        # hdulist = pyfits.HDUList([hdu])
        # hdulist.writeto(outputdir+os.path.sep+'fmout1_mvt6_KL30.fits',clobber=True)

        # The mf.MatchedFilter class calculate the projection of the FM on the data for each pixel and images.
        # The final combination to form the cross  cross correlation, matched filter and contrast maps is done right
        # here.
        FMCC_map = np.nansum(fmout1[0,:,:,:,:,:],axis=2) \
                        / np.sqrt(np.nansum(fmout1[1,:,:,:,:,:],axis=2))
        FMCC_map[np.where(FMCC_map==0)]=np.nan
        self.FMCC_map = FMCC_map

        FMMF_map = np.nansum(fmout1[0,:,:,:,:,:]/fmout1[2,:,:,:,:,:],axis=2) \
                        / np.sqrt(np.nansum(fmout1[1,:,:,:,:,:]/fmout1[2,:,:,:,:,:],axis=2))
        FMMF_map[np.where(FMMF_map==0)]=np.nan
        self.FMMF_map = FMMF_map

        contrast_map = np.nansum(fmout1[0,:,:,:,:,:]/fmout1[2,:,:,:,:,:],axis=2) \
                        / np.nansum(fmout1[1,:,:,:,:,:]/fmout1[2,:,:,:,:,:],axis=2)
        contrast_map[np.where(contrast_map==0)]=np.nan
        self.contrast_map = contrast_map

        self.FMNpix_map = np.nansum(fmout1[3,:,:,:,:,:],axis=2)

        self.metricMap = [self.FMMF_map,self.FMCC_map,self.contrast_map]


        for k in range(np.size(self.numbasis)):
            # Save the outputs (matched filter, shape map and klipped image) as fits files
            suffix = "FMMF-KL{0}".format(self.numbasis[k])
            dataset.savedata(outputdir+os.path.sep+fileprefix+'-'+suffix+'.fits',
                             self.FMMF_map[0,k,:,:],
                             filetype=suffix)

            suffix = "FMCont-KL{0}".format(self.numbasis[k])
            dataset.savedata(outputdir+os.path.sep+fileprefix+'-'+suffix+'.fits',
                             self.contrast_map[0,k,:,:],
                             filetype=suffix)

            suffix = "FMCC-KL{0}".format(self.numbasis[k])
            dataset.savedata(outputdir+os.path.sep+fileprefix+'-'+suffix+'.fits',
                             self.FMCC_map[0,k,:,:],
                             filetype=suffix)

            suffix = "FMN_pix-KL{0}".format(self.numbasis[k])
            dataset.savedata(outputdir+os.path.sep+fileprefix+'-'+suffix+'.fits',
                             self.FMNpix_map[0,k,:,:],
                             filetype=suffix)

            if self.save_bbfm:
                suffix = "BBFM-KL{0}".format(self.numbasis[k])
                dataset.savedata(outputdir+os.path.sep+fileprefix+'-'+suffix+'.fits',
                                 fmout2,
                                 filetype=suffix)
            elif self.save_fm:
                suffix = "AllFM-KL{0}".format(self.numbasis[k])
                dataset.savedata(outputdir+os.path.sep+fileprefix+'-'+suffix+'.fits',
                                 fmout2,
                                 filetype=suffix)

        return

    def generate_model_sci(self, input_img_shape, section_ind, pa, wv, radstart, radend, phistart, phiend, padding, ref_center, parang, ref_wv,sep_fk,pa_fk, flipx):
        """
        Generate model PSFs at the correct location of this segment of the science image denotated by its wv and
        parallactic angle.

        Args:
            input_img_shape: 2-D shape of inpt images ([ysize, xsize])
            section_ind: array indicies into the 2-D x-y image that correspond to this section.
                         Note needs be called as section_ind[0]
            pa: parallactic angle of the science image [degrees]
            wv: wavelength of the science image
            radstart: radius of start of segment (not used)
            radend: radius of end of segment (not used)
            phistart: azimuthal start of segment [radians] (not used)
            phiend: azimuthal end of segment [radians] (not used)
            padding: amount of padding on each side of sector
            ref_center: center of image
            parang: parallactic angle of input image [DEGREES] (not used)
            ref_wv: wavelength of science image
            sep_fk: separation of the planet to be injected.
            pa_fk: position angle of the planet to be injected.
            flipx: if True, flip x coordinate in final image

        Return: (models, mask)
            models: vector of size p where p is the number of pixels in the segment
            mask: vector of size p where p is the number of pixels in the segment
                    if pixel == 1: arc shape where to calculate the standard deviation
                    if pixel == 2: 7 pixels disk around the position of the planet.
        """
        # create some parameters for a blank canvas to draw psfs on
        nx = input_img_shape[1]
        ny = input_img_shape[0]

        if hasattr(self,"x_grid") and hasattr(self,"y_grid"):
            x_grid, y_grid = self.x_grid,self.y_grid
        else:
            x_grid, y_grid = np.meshgrid(np.arange(nx * 1.)-ref_center[0], np.arange(ny * 1.)-ref_center[1])

        numwv, ny_psf, nx_psf =  self.input_psfs.shape

        # a blank img array of write model PSFs into
        whiteboard = np.zeros((ny,nx))
        # grab PSF given wavelength
        wv_index = spec.find_nearest(self.input_psfs_wvs,wv)[1]

        sign = -1.
        if flipx:
            sign = 1.

        # The trigonometric calculation are save in a dictionary to avoid calculating them many times.
        recalculate_trig = False
        if pa not in self.psf_centx_notscaled:
            recalculate_trig = True
        else:
            if pa_fk != self.curr_pa_fk[pa] or sep_fk != self.curr_sep_fk[pa]:
                recalculate_trig = True
        if recalculate_trig: # we could actually store the values for the different pas too...
            # flipx requires the opposite rotation
            self.psf_centx_notscaled[pa] = sep_fk * np.cos(np.radians(90. - sign*pa_fk - pa))
            self.psf_centy_notscaled[pa] = sep_fk * np.sin(np.radians(90. - sign*pa_fk - pa))
            self.curr_pa_fk[pa] = pa_fk
            self.curr_sep_fk[pa] = sep_fk

        psf_centx = (ref_wv/wv) * self.psf_centx_notscaled[pa]
        psf_centy = (ref_wv/wv) * self.psf_centy_notscaled[pa]

        # create a coordinate system for the image that is with respect to the model PSF
        # round to nearest pixel and add offset for center
        l = int(round(psf_centx + ref_center[0]))
        k = int(round(psf_centy + ref_center[1]))
        # recenter coordinate system about the location of the planet
        # x_vec_stamp_centered = x_grid[0, (l-col_m):(l+col_p)]-psf_centx
        # y_vec_stamp_centered = y_grid[(k-row_m):(k+row_p), 0]-psf_centy
        x_vec_stamp_centered = x_grid[0, np.max([(l-self.col_m),0]):np.min([(l+self.col_p),nx])]-psf_centx
        y_vec_stamp_centered = y_grid[np.max([(k-self.row_m),0]):np.min([(k+self.row_p),ny]), 0]-psf_centy
        # rescale to account for the align and scaling of the refernce PSFs
        # e.g. for longer wvs, the PSF has shrunk, so we need to shrink the coordinate system
        x_vec_stamp_centered /= (ref_wv/wv)
        y_vec_stamp_centered /= (ref_wv/wv)

        # use intepolation spline to generate a model PSF and write to temp img
        # whiteboard[(k-row_m):(k+row_p), (l-col_m):(l+col_p)] = \
        #         self.psfs_func_list[wv_index[0]](x_vec_stamp_centered,y_vec_stamp_centered).transpose()
        whiteboard[np.max([(k-self.row_m),0]):np.min([(k+self.row_p),ny]), np.max([(l-self.col_m),0]):np.min([(l+self.col_p),nx])] = \
                self.psfs_func_list[wv_index](x_vec_stamp_centered,y_vec_stamp_centered).transpose()

        # write model img to output (segment is collapsed in x/y so need to reshape)
        whiteboard.shape = [input_img_shape[0] * input_img_shape[1]]
        segment_with_model = copy(whiteboard[section_ind])
        whiteboard.shape = [input_img_shape[0],input_img_shape[1]]

        # Define the masks for where the planet is and the background.
        if hasattr(self,"th0_grid") and hasattr(self,"r_grid"):
            r_grid = self.r_grid
            th_grid = (self.th0_grid-sign*np.radians(pa))% (2.0 * np.pi)
        else:
            r_grid = abs(x_grid +y_grid*1j)
            th_grid = (np.arctan2(sign*x_grid,y_grid)-sign*np.radians(pa))% (2.0 * np.pi)
        w = self.background_width
        thstart = (np.radians(pa_fk)- float(w)/sep_fk) % (2.0 * np.pi) # -(2*np.pi-np.radians(pa))
        thend = (np.radians(pa_fk) + float(w)/sep_fk) % (2.0 * np.pi) # -(2*np.pi-np.radians(pa))
        # thstart = (np.radians(pa_fk)- 2*float(w)/sep_fk) % (2.0 * np.pi) # -(2*np.pi-np.radians(pa))
        #thend = (np.radians(pa_fk) + 2*float(w)/sep_fk) % (2.0 * np.pi) # -(2*np.pi-np.radians(pa))
        if thstart < thend:
            where_mask = np.where((r_grid>=(sep_fk-w)) & (r_grid<(sep_fk+w)) & (th_grid >= thstart) & (th_grid < thend))
        else:
            where_mask = np.where((r_grid>=(sep_fk-w)) & (r_grid<(sep_fk+w)) & ((th_grid >= thstart) | (th_grid < thend)))
        whiteboard[where_mask] = 1
        #TODO check the modification I did to these lines
        whiteboard = np.pad(whiteboard,((self.row_m,self.row_p),(self.col_m,self.col_p)),mode="constant",constant_values=0)
        # whiteboard[(k-row_m):(k+row_p), (l-col_m):(l+col_p)][np.where(np.isnan(self.stamp_PSF_mask))]=2
        whiteboard[(k):(k+self.row_m+self.row_p), (l):(l+self.col_m+self.col_p)][np.where(np.isnan(self.stamp_PSF_mask))]=2
        whiteboard = np.ascontiguousarray(whiteboard[self.row_m:self.row_m+input_img_shape[0],self.col_m:self.col_m+input_img_shape[1]])
        whiteboard.shape = [input_img_shape[0] * input_img_shape[1]]
        mask = whiteboard[section_ind]

        # create a canvas to place the new PSF in the sector on
        if 0:#np.size(np.where(mask==2)[0])==0: 296
            print(pa,pa_fk)
            print(thstart,thend)
            whiteboard[section_ind] = whiteboard[section_ind] + 0.5
            whiteboard.shape = (input_img_shape[0], input_img_shape[1])
            blackboard = np.zeros((ny,nx))
            blackboard.shape = [input_img_shape[0] * input_img_shape[1]]
            blackboard[section_ind] = segment_with_model
            blackboard.shape = [input_img_shape[0],input_img_shape[1]]
            plt.figure(1)
            plt.subplot(1,3,1)
            im = plt.imshow(whiteboard)
            plt.colorbar(im)
            plt.subplot(1,3,2)
            im = plt.imshow(blackboard)
            plt.colorbar(im)
            plt.subplot(1,3,3)
            im = plt.imshow(np.degrees(th_grid))
            plt.colorbar(im)
            plt.show()

        return segment_with_model,mask

    def generate_models(self, input_img_shape, section_ind, pas, wvs, radstart, radend, phistart, phiend, padding, ref_center, parang, ref_wv,sep_fk,pa_fk,flipx):
        """
        Generate model PSFs at the correct location of this segment for each image denotated by its wv and parallactic
        angle.

        Args:
            input_img_shape: 2-D shape of inpt images ([ysize, xsize])
            section_ind: array indicies into the 2-D x-y image that correspond to this section.
                         Note needs be called as section_ind[0]
            pas: array of N parallactic angles corresponding to N images [degrees]
            wvs: array of N wavelengths of those images
            radstart: radius of start of segment (not used)
            radend: radius of end of segment (not used)
            phistart: azimuthal start of segment [radians] (not used)
            phiend: azimuthal end of segment [radians] (not used)
            padding: amount of padding on each side of sector
            ref_center: center of image
            parang: parallactic angle of input image [DEGREES] (not used)
            ref_wv: wavelength of science image
            sep_fk: separation of the planet to be injected.
            pa_fk: position angle of the planet to be injected.
            flipx: if True, flip x coordinate in final image

        Return:
            models: array of size (N, p) where p is the number of pixels in the segment
        """
        # create some parameters for a blank canvas to draw psfs on
        nx = input_img_shape[1]
        ny = input_img_shape[0]
        try:
            x_grid, y_grid = self.x_grid,self.y_grid
        except:
            x_grid, y_grid = np.meshgrid(np.arange(nx * 1.)-ref_center[0], np.arange(ny * 1.)-ref_center[1])

        sign = -1.
        if flipx:
            sign = 1.

        # a blank img array of write model PSFs into
        whiteboard = np.zeros((ny,nx))
        models = []
        #print(self.input_psfs.shape)
        for pa, wv in zip(pas, wvs):
            # grab PSF given wavelength
            wv_index = [spec.find_nearest(self.input_psfs_wvs,wv)[1]]

            # find center of psf
            # to reduce calculation of sin and cos, see if it has already been calculated before
            recalculate_trig = False
            if pa not in self.psf_centx_notscaled:
                recalculate_trig = True
            else:
                #print(self.psf_centx_notscaled[pa],pa)
                if pa_fk != self.curr_pa_fk[pa] or sep_fk != self.curr_sep_fk[pa]:
                    recalculate_trig = True
            if recalculate_trig: # we could actually store the values for the different pas too...
                self.psf_centx_notscaled[pa] = sep_fk * np.cos(np.radians(90. - sign*pa_fk - pa))
                self.psf_centy_notscaled[pa] = sep_fk * np.sin(np.radians(90. - sign*pa_fk - pa))
                self.curr_pa_fk[pa] = pa_fk
                self.curr_sep_fk[pa] = sep_fk

            psf_centx = (ref_wv/wv) * self.psf_centx_notscaled[pa]
            psf_centy = (ref_wv/wv) * self.psf_centy_notscaled[pa]

            # create a coordinate system for the image that is with respect to the model PSF
            # round to nearest pixel and add offset for center
            l = int(round(psf_centx + ref_center[0]))
            k = int(round(psf_centy + ref_center[1]))
            # recenter coordinate system about the location of the planet
            x_vec_stamp_centered = x_grid[0, (l-self.col_m):(l+self.col_p)]-psf_centx
            y_vec_stamp_centered = y_grid[(k-self.row_m):(k+self.row_p), 0]-psf_centy
            # rescale to account for the align and scaling of the refernce PSFs
            # e.g. for longer wvs, the PSF has shrunk, so we need to shrink the coordinate system
            x_vec_stamp_centered /= (ref_wv/wv)
            y_vec_stamp_centered /= (ref_wv/wv)

            # use intepolation spline to generate a model PSF and write to temp img
            whiteboard[(k-self.row_m):(k+self.row_p), (l-self.col_m):(l+self.col_p)] = \
                    self.psfs_func_list[wv_index[0]](x_vec_stamp_centered,y_vec_stamp_centered).transpose()

            # write model img to output (segment is collapsed in x/y so need to reshape)
            whiteboard.shape = [input_img_shape[0] * input_img_shape[1]]
            segment_with_model = copy(whiteboard[section_ind])
            whiteboard.shape = [input_img_shape[0],input_img_shape[1]]

            models.append(segment_with_model)

            # create a canvas to place the new PSF in the sector on
            if 0:
                blackboard = np.zeros((ny,nx))
                blackboard.shape = [input_img_shape[0] * input_img_shape[1]]
                blackboard[section_ind] = segment_with_model
                blackboard.shape = [input_img_shape[0],input_img_shape[1]]
                plt.figure(1)
                plt.subplot(1,2,1)
                im = plt.imshow(whiteboard)
                plt.colorbar(im)
                plt.subplot(1,2,2)
                im = plt.imshow(blackboard)
                plt.colorbar(im)
                plt.show()

            whiteboard[(k-self.row_m):(k+self.row_p), (l-self.col_m):(l+self.col_p)] = 0.0

        return np.array(models)



def calculate_fm_opti(delta_KL, original_KL, sci, model_sci_fk,delta_KL_fk, original_KL_fk):
    r"""
    Optimized version for calculate_fm() (if numbas) for a single numbasis.

    Calculate what the PSF looks up post-KLIP using knowledge of the input PSF, assumed spectrum of the science target,
    and the partially calculated KL modes (\Delta Z_k^\lambda in Laurent's paper). If inputflux is None,
    the spectral dependence has already been folded into delta_KL_nospec (treat it as delta_KL).

    Note: if inputflux is None and delta_KL_nospec has three dimensions (ie delta_KL_nospec was calculated using
    pertrurb_nospec() or perturb_nospec_modelsBased()) then only klipped_oversub and klipped_selfsub are returned.
    Besides they will have an extra first spectral dimension.

    Args:
        delta_KL_nospec: perturbed KL modes but without the spectral info. delta_KL = spectrum x delta_Kl_nospec.
                         Shape is (numKL, wv, pix). If inputflux is None, delta_KL_nospec = delta_KL
        orignal_KL: unpertrubed KL modes (array of size [numbasis, numpix])
        sci: array of size p representing the science data
        model_sci: array of size p corresponding to the PSF of the science frame
        input_spectrum: array of size wv with the assumed spectrum of the model

    If delta_KL_nospec does NOT include a spectral dimension or if inputflux is not None:
    Returns:
        fm_psf: array of shape (b,p) showing the forward modelled PSF
                Skipped if inputflux = None, and delta_KL_nospec has 3 dimensions.
        klipped_oversub: array of shape (b, p) showing the effect of oversubtraction as a function of KL modes
        klipped_selfsub: array of shape (b, p) showing the effect of selfsubtraction as a function of KL modes
        Note: psf_FM = model_sci - klipped_oversub - klipped_selfsub to get the FM psf as a function of K Lmodes
              (shape of b,p)

    If inputflux = None and if delta_KL_nospec include a spectral dimension:
    Returns:
        klipped_oversub: Sum(<S|KL>KL) with klipped_oversub.shape = (1,Npix)
        klipped_selfsub: Sum(<N|DKL>KL) + Sum(<N|KL>DKL) with klipped_selfsub.shape = (1,N_lambda or N_ref,N_pix)

    """
    # remove means and nans from science image
    sci_mean_sub = (sci - np.nanmean(sci))[None,:]
    sci_mean_sub[np.where(np.isnan(sci_mean_sub))] =0


    # science PSF models, ready for FM
    # /!\ JB: If subtracting the mean. It should be done here. not in klip_math since we don't use model_sci there.
    model_sci_mean_sub = model_sci_fk[None,:] # should be subtracting off the mean?
    model_sci_mean_sub[np.where(np.isnan(model_sci_mean_sub))] =0


    # Forward model the PSF
    # 3 terms: 1 for oversubtracton (planet attenauted by speckle KL modes),
    # and 2 terms for self subtraction (planet signal leaks in KL modes which get projected onto speckles)
    #
    # Klipped = N-Sum(<N|KL>KL) + S-Sum(<S|KL>KL) - Sum(<N|DKL>KL) - Sum(<N|KL>DKL)
    # With  N = noise/speckles (science image)
    #       S = signal/planet model
    #       KL = KL modes
    #       DKL = perturbation of the KL modes/Delta_KL
    #
    # sci_mean_sub_rows.shape = (1,N_pix)
    # model_sci_mean_sub_rows.shape = (1,N_pix)
    # original_KL.shape = (max_basis,N_pix)
    # delta_KL.shape = (max_basis,N_pix)
    oversubtraction_inner_products = np.dot(model_sci_mean_sub, original_KL_fk.T)
    selfsubtraction_1_inner_products = np.dot(sci_mean_sub, delta_KL.T)
    selfsubtraction_2_inner_products = np.dot(sci_mean_sub, original_KL.T)

    # oversubtraction_inner_products = (1,numbasis)
    klipped_oversub = np.dot(oversubtraction_inner_products, original_KL_fk)

    # selfsubtraction_1_inner_products = (1,numbasis)
    # selfsubtraction_2_inner_products = (1,numbasis)
    klipped_selfsub = np.dot(selfsubtraction_1_inner_products, original_KL_fk) + \
                      np.dot(selfsubtraction_2_inner_products, delta_KL_fk)

    return model_sci_fk[None,:] - klipped_oversub - klipped_selfsub


#!/usr/bin/env python

"""
This library has method for operating on P1640 core images
"""

import sys
import os
import glob
import re

import numpy as np
import pandas as pd
from scipy import interpolate

from astropy.io import fits
from photutils import aperture_photometry, CircularAperture

sys.path.append("~/pyklip")
from pyklip.klip import align_and_scale

def get_cube_xsection(orig_cube, center, width):
    """
    Select the cross-section of a cube centered in center with 1/2-width width
    Input:
        orig_cube: Nlambda x Npix x Npix datacube
        center: [row, col] index
        width: 1/2-width of cross-section
    Returns:
       cube_cutout: Nlambda x (2*width+1) x (2*width+1) cube cross-section
    """
    cube = orig_cube.copy()
    shape = cube.shape
    try:
        assert(np.all(np.where(center < 0)))
        assert(center[0] < shape[1])
        assert(center[1] < shape[2])
    except AssertionError:
        print("bad value for center")
        return None
    # [(xlow, xhigh),(ylow,yhigh)]
    xlims = (np.floor(np.nanmax([0,center[0]-width])).astype(np.int),
             np.ceil(np.nanmin([shape[-1],center[0]+width])).astype(np.int))
    ylims = (np.floor(np.nanmax([0,center[1]-width])).astype(np.int),
             np.ceil(np.nanmin([shape[-1],center[1]+width])).astype(np.int))
    center_lims = np.array([xlims, ylims])
    cube_cutout = cube[:, 
                       center_lims[0][0]:center_lims[0][1], 
                       center_lims[1][0]:center_lims[1][1]].copy()
    return cube_cutout


def centroid_image(orig_img):
    """
    Centroid an image - weighted sum of pixels
    Input:
        orig_img: 2D array
    Return:
        [y,x] floating-point coordinates of the centroid
    """
    img = orig_img.copy()
    # don't include negative pixels; they are not part of the PSF
    img[img<0] = 0.
    ypix, xpix = np.indices(img.shape)
    tot = np.nansum(img)
    ycenter = np.dot(np.ravel(ypix), np.ravel(img))/tot
    xcenter = np.dot(np.ravel(xpix), np.ravel(img))/tot
    return ycenter, xcenter

def get_PSF_center(orig_cube, refchan=26, fine=False):
    """
    Return the PSF center at the pixel level (default) or subpixel level (fine=True)
    Input:
        orig_cube: [Nlambda x] Npix x Npix datacube (or image)
        refchan(=26): Reference channel for the initial center estimate
        fine_centering(=False): After getting a rough estimate of the center, centroid the image
    Returns:
        Nlamdba x 2 array of pixel indices for the PSF center
    """
    cube = orig_cube.copy()
    orig_shape = cube.shape
    IMG_FLAG = True if len(cube.shape) == 2 else False
    if IMG_FLAG:
        cube = cube[np.newaxis,:,:]
    shape = cube.shape
    nchan = shape[0]
    if nchan < refchan:
        refchan = nchan-1
    init_center = np.array(np.unravel_index(np.nanargmax(cube[refchan]), shape[-2:]))
    width = 15
    center_cutout = get_cube_xsection(cube, init_center, width)
    try:
        centers = np.array([np.unravel_index(np.nanargmax(center_cutout[chan]), center_cutout[chan].shape)
                            for chan in range(nchan)] + init_center - width)
    except ValueError:
        print("All-NaN slice encountered:", chan)
        return None
    # If you want centroiding, do a second pass on a cross-section of the cube centered on the initial guess
    if fine == True:
        init_center = np.floor(np.nanmean(centers,axis=0)).astype(np.int)
        center_cutout = get_cube_xsection(cube, init_center, width)
        centers = np.array([centroid_image(img) for img in center_cutout]) + init_center - width
    # negative values are illegal, set to bottom corner
    centers = np.fmax(0, centers)
    
    if IMG_FLAG: # take care of dimensionality
        return centers[0]
    return centers


def _get_encircled_energy_image(im, center, frac=0.5):
    """
    Get the encircled energy for a 2-D image.
    Calculate the radial pixel coordinates centered on 'center'. 
    """
    yx_centered = np.array((np.indices(im.shape).T - center).T)
    # convert to radii
    rad_coord = np.linalg.norm(yx_centered, axis=0)

    radpix = pd.DataFrame(zip(np.ravel(rad_coord), np.ravel(im)), 
                          columns=['rad', 'pixval'])
    radpix.sort(columns='rad', inplace=True)
    radpix.reset_index(inplace=True)
    radpix['cumsum'] = radpix['pixval'].cumsum()
    unique_rad = np.unique(radpix['rad'])
    # group by common radius
    grouped = radpix.groupby('rad') 
    unique_flux = grouped.pixval.sum().as_matrix()
    pix_per_rad = grouped.size().as_matrix()

    unique_cumsum = np.cumsum(unique_flux)
    psf_edge_pix = np.nanargmax(unique_cumsum) + 1 # +1 is added for array indexing reasons
    psf_edge_rad = unique_rad[psf_edge_pix-1]
    bgnd_mean = np.nanmean(unique_flux[psf_edge_pix:])
    bgnd_std = np.nanstd(radpix[radpix['rad'] > psf_edge_rad]['pixval'])
    # subtract background
    unique_flux -= bgnd_mean*pix_per_rad
    # redo cumsum now with bgnd subtracted
    unique_cumsum = np.cumsum(unique_flux)
    total_flux = unique_cumsum[psf_edge_pix-1]
    encircled_energy = unique_cumsum[:psf_edge_pix]/total_flux

    ind = ['starx','stary','radius','flux','bgnd_mean','bgnd_std', 'bgnd_npix']
    ee_data = pd.Series(np.zeros(len(ind)), index=ind)
    ee_data[['stary','starx']] = center
    ee_data['flux'] = total_flux
    ee_data['bgnd_mean'] = bgnd_mean
    ee_data['bgnd_std'] = bgnd_std
    ee_data['bgnd_npix'] = grouped.size().ix[psf_edge_rad:].sum()

    # get the radius, taking care of the case of a crappy channel
    try:
        lolim = np.where(encircled_energy < frac)[0][-1]
        hilim = np.where(encircled_energy >= frac)[0][0]
    except IndexError:
        ee_data['radius'] = 1
        return ee_data
    encirc_interp = interpolate.interp1d(x=encircled_energy[[lolim, hilim]], 
                                         y=unique_rad[[lolim, hilim]])
    ee_data['radius'] = encirc_interp(frac)

    return ee_data


def get_encircled_energy_cube(orig_cube, frac=0.5):
    """
    Get the fractional encircled energy of a PSF in each channel of a datacube. 
    Basically a wrapper for _get_encircled_energy_image. Accepts 2-D and 3-D input.
    Input:
        orig_cube: unocculted core cube [Nlambda x ]Npix x Npix
        frac: encircled energy cutoff
    Returns:
        Pandas Dataframe with Nlambda columns:
        [starx, stary, radius, flux]
    """
    cube = orig_cube.copy()
    if len(cube.shape) == 2:
        cube = cube[np.newaxis,:,:]
    nchan = cube.shape[0]
    centers = get_PSF_center(cube, fine=True)
    ee_data = pd.DataFrame([_get_encircled_energy_image(im, center, frac)
                            for im, center in zip(cube, centers)])
    return ee_data


def _aperture_convolve_img(orig_img, aperture_radius, apkwargs={'method':'subpixel','subpixels':4}):
    """
    Perform aperture photometry on every pixel in a 2-D image
    Input:
        orig_img: Npix x Npix image
        aperture_radius: radius of the circular aperture
    Returns:
        phot_img: image of the aperture photometry on every non-nan pixel
    """
    phot_img = np.zeros(orig_img.shape)
    nanpix = np.where(np.isnan(orig_img))
    notnanpix = np.where(~np.isnan(orig_img))

    orig_shape = orig_img.shape
    img = orig_img.copy() # let's not modify the original cube
    img[nanpix] = 0.0
    
    # put an aperture on every non-nan pixel
    img_apertures = CircularAperture(positions = zip(*notnanpix[::-1]),
                                     r = aperture_radius)
    img_phot = aperture_photometry(img, img_apertures, **apkwargs)
    # set up to calculate the flattened indices
    ncols = np.ones_like(img_phot['ycenter'])*img.shape[-1]
    flattened_ind = lambda n, row, col: n*row+col
    ind = np.array(map(flattened_ind, ncols,
                       img_phot['ycenter'],
                       img_phot['xcenter']),
                   dtype=np.int)
    phot_img = np.ravel(phot_img)
    # map the photometry to the cube
    phot_img[ind] = img_phot['aperture_sum']
    phot_img = phot_img.reshape(orig_shape)
    # re-assign the nan pixels
    phot_img[nanpix] = np.nan
    
    return phot_img


def aperture_convolve_cube(orig_cube, aperture_radii, apkwargs={'method':'subpixel','subpixels':4}):
    """
    Perform apeture photometry on every pixel in a datacube or image
    Wrapper for _aperture_convolve_img to handle 2-D and 3-D data
    Input:
        orig_cube: [Nlambda x] Npix x Npix datacube
        aperture_radii: Nlambda array of aperture radii
        apkwargs: dictionary of arguments to pass to aperture_photometry
            Default: {'method':'subpixel','subpixels':4}
    Returns:
        phot_cube: cube with shape of orig_cube of the aperture photometry
    """
    cube = orig_cube.copy()
    orig_shape = cube.shape
    aperture_radii = np.array(aperture_radii) # because pandas indexing can cause problems
    # image or cube?
    IMG_FLAG = True if len(orig_shape) == 2 else False
    if len(orig_shape) == 2:
        cube = cube[np.newaxis,:,:]
    phot_cube = np.zeros(cube.shape)
    for i, im in enumerate(cube):
        phot_cube[i] = _aperture_convolve_img(im, aperture_radii[i], apkwargs)

    if IMG_FLAG:
        return phot_cube[0]
    return phot_cube


def combine_multiple_cores(multiple_core_info):
    """
    Combine the stellar flux and radii information from multiple cores in the proper way.
    """
    rad_mean = np.nanmean([c['radius'] for c in multiple_core_info], axis=0)
    flux_mean = np.nanmean([c['flux'] for c in multiple_core_info], axis=0)
    rad_err = np.nanstd([c['radius'] for c in multiple_core_info], axis=0)
    flux_err = np.nanstd([c['flux'] for c in multiple_core_info], axis=0)

    core_info = pd.DataFrame(zip(rad_mean, rad_err, flux_mean, flux_err),
                             columns=['radius','drad','flux','dflux'])
    return core_info

def zero_pad_core_box(core_cutout, centers, radii):
    """
    Get a cube of core cutouts with a center and a radius for each channel
    Inside/outside the radius is determined by the center of the pixel.
    Negative pixels are set to 0.
    """
    pass
    

def get_injection_core(core_cubes):
    """
    Remember, the injected PSF needs to be the SAME as the reference PSF, except for a scaling factor!
    Combine multiple cubes into a single core file for injection.
    Make sure that the total injected flux is the sum of the pixels!
    Outline:
    1. Get the encircled energy fraction for all the cores (frac=1)
    2. For each core, prepare a zero-cube with width of the largest radius + 1
    3. Add the core from each channel to the zero-cube
    
    """
    core_info_all = [get_encircled_energy_cube(c, frac=1) for c in core_cubes]
    core_cutouts = []
    for cube, core_df in zip(core_cubes, core_info_all):
        core_cutouts.append(get_core_xsection(cube,
                                        core_df[['stary','starx']].mean(), 
                                        core['radius'].max()))
        # set it so the cutout sum gives the total flux in each channel
        core_cutouts[-1][core_cutout<0] = 0
        core_cutouts[-1][np.where(np.isnan(core_cutout))] = 0

    core_cutouts = np.array(core_cutouts)
    final_psf = np.median(core_cutouts, axis=0)
    return final_psf

def make_median_core(core_cubes):
    """
    Take a set of core cubes and assemble a median cube out of them. Set all non-PSF pixels to 0
    Input:
        core_cubes: Ncubes x Nlambda x Npix x Npix set of cores
    Returns:
        median_core: Nlambda x Npix x Npix 
    """
    """
    Pseudocode:
    For each core:
        1. Get full radius for each channel in each core.
        2. Make an aligned cutout
        3. Mask all non-PSF pixels
        4. Combine the cutouts into an array of length core_cubes
    Median-combine the unmasked pixels
    """
    core_dfs = [get_encircled_energy_cube(cube, frac=1.) for cube in core_cubes]
    tmp_core_cutouts = [get_cube_xsection(core_cubes[i],
                                          df[['stary','starx']].mean(),
                                          df['radius'].max()+i)
                        for i,df in enumerate(core_dfs)]
    # make sure all the core cutouts have the same dimensions
    # if the radii are different, the difference in dims is 2 pixels
    max_shape = np.max([c.shape for c in tmp_core_cutouts], axis=0)
    core_cutouts = np.zeros((len(tmp_core_cutouts), max_shape[-3],
                             max_shape[-2], max_shape[-1]))*np.nan # make a nan-cube
    for i, cutout in enumerate(tmp_core_cutouts):
        shift = np.abs((cutout.shape[-2:] - max_shape[-2:])/2).astype(np.int)
        core_cutouts[i,:,shift[-2]:max_shape[-2]-shift[-2],shift[-1]:max_shape[-1]-shift[-1]] = cutout[:,:,:].copy()
    # align the cutouts so you can median-combine them
    # only need to align corresponding channels
    aligned_cutouts = core_cutouts.copy()
    dims = aligned_cutouts.shape
    aligned_cutouts = aligned_cutouts.reshape(reduce(lambda x,y:x*y,dims[:-2]), dims[-2], dims[-1])
    new_center = map(lambda x: np.floor(x/2.).astype(np.int), core_cutouts.shape[-2:])[::-1]
    old_centers = get_PSF_center(aligned_cutouts, fine=True)[:,::-1]
    for i in range(len(aligned_cutouts)):
        aligned_cutouts[i] = align_and_scale(aligned_cutouts[i], new_center, old_centers[i])
    aligned_cutouts = aligned_cutouts.reshape(dims)

    # before subtracting off the background, make a mask to ID background pixels
    bgnd_mask = np.ma.mask_or((aligned_cutouts < 0), np.isnan(aligned_cutouts))
    # before median-combining, subtract off the background, and:
    background = np.array([df['bgnd_mean'] for df in core_dfs])
    cutout_psfs = np.ma.masked_array([(psf.T - bg).T for psf,bg in zip(aligned_cutouts, background)],
                                     mask=bgnd_mask,
                                     fill_value=0.0)
    median_psf = np.ma.median(cutout_psfs, axis=0)
    # return an unmasked array
    return median_psf.data
    



__author__ = 'jruffio'

import numpy as np
from scipy.interpolate import interp1d
import astropy.io.fits as pyfits
import platform
import scipy.ndimage as ndimage
from scipy.optimize import leastsq
from scipy.optimize import minimize
import glob
import os
import csv




def find_upper_nearest(array,value):
    """
    Find the upper nearest element to value in array.

    :param array: Array of value
    :param value: Value for which one wants the upper value.
    :return: (up_value, id) with up_value the closest upper value and id its index.
    """
    diff = array-value
    diff[np.where(diff<0.0)] = np.nan
    idx = np.nanargmin(diff)
    return array[idx], idx

def find_lower_nearest(array,value):
    """
    Find the lower nearest element to value in array.

    :param array: Array of value
    :param value: Value for which one wants the lower value.
    :return: (low_value, id) with low_value the closest lower value and id its index.
    """
    diff = array-value
    diff[np.where(diff>0.0)] = np.nan
    idx = np.nanargmax(diff)
    return array[idx], idx


def find_nearest(array,value):
    """
    Find the nearest element to value in array.

    :param array: Array of value
    :param value: Value for which one wants the closest value.
    :return: (closest_value, id) with closest_value the closest lower value and id its index.
    """
    diff = np.array(array)-value
    # diff[np.where(diff>0.0)] = np.nan
    idx = np.nanargmin(np.abs(diff))
    return array[idx], idx

def get_specType(object_name,SpT_file_csv = None):
    """
    Return the spectral type for a target based on Simbad or on the table in SpT_file

    :param object_name: Name of the target: ie "c_Eri"
    :param SpT_file: Filename (.csv) of the table containing the target names and their spectral type.
                    Can be generated from quering Simbad.
                    If None (default), the function directly tries to query Simbad.
    :return: Spectral type
    """
    # Hard-coded spectral type for some targets. Not ideal but I don't want to think about it right now.
    if object_name == "iot_Cen":
        return "A1"
    if object_name == "IK_Peg":
        return "A8"

    if SpT_file_csv is not None:
        with open(SpT_file_csv, 'r') as csvfile_TID:
            TID_reader = csv.reader(csvfile_TID, delimiter=';')
            TID_csv_as_list = list(TID_reader)
            TID_csv_as_nparr = np.array(TID_csv_as_list)[1:len(TID_csv_as_list),:]
            target_names = np.ndarray.tolist(TID_csv_as_nparr[:,0])
            specTypes = np.ndarray.tolist(TID_csv_as_nparr[:,1])

    try:
        object_name = object_name.replace(' ','_')
        object_name = object_name.replace('+','_')
        return specTypes[target_names.index(object_name)]
    except:
        object_name = object_name.replace('_','+')
        object_name = object_name.replace(' ','+')

        try: #python 2
            import urllib
            # url = urllib.urlopen("http://cdsweb.u-strasbg.fr/cgi-bin/nph-sesame/-oI?"+object_name)
            url = urllib.urlopen("http://simbad.u-strasbg.fr/simbad/sim-id?output.format=ASCII&output.max=1&\
                                  obj.cooN=off&obj.pmsel=off&obj.plxsel=off&obj.rvsel=off&obj.spsel=on&obj.mtsel=off&\
                                  obj.sizesel=off&obj.fluxsel=off&obj.messel=off&obj.notesel=off&obj.bibsel=off&Ident="+object_name)
        except: # python 3
            import urllib.request
            # url = urllib.request.urlopen("http://www.python.org")
            url = urllib.request.urlopen("http://simbad.u-strasbg.fr/simbad/sim-id?output.format=ASCII&output.max=1&"
                                        +"obj.cooN=off&obj.pmsel=off&obj.plxsel=off&obj.rvsel=off&obj.spsel=on&obj.mtsel=off&"
                                        +"obj.sizesel=off&obj.fluxsel=off&obj.messel=off&obj.notesel=off&obj.bibsel=off&Ident="+object_name)
        text = url.read()
        for line in text.splitlines():
            str_line = line.decode("utf8")
            # if line.find('Spectral type:') != -1:
            if 'Spectral type:' in str_line:
                # print(line)
                spec_type =str_line.split("Spectral type: ")[-1].replace("Spectral type: ","").split(" ")[0]

        if SpT_file_csv is not None:
            print("Couldn't find {0} in spectral type list. Retrieved it from Simbad.".format(object_name))
            target_names.append(object_name.replace('+','_'))
            specTypes.append(spec_type)
            attrib_name=["name","specType"]
            with open(SpT_file_csv, 'w+') as csvfile:
                csvwriter = csv.writer(csvfile, delimiter=';')
                table_to_csv = [attrib_name]+[(name, stype) for name, stype in zip(target_names,specTypes)]
                csvwriter.writerows(table_to_csv)

        return spec_type

def get_star_spectrum(wvs_or_filter_name,star_type = None, temperature = None,mute = None):
    """
    Get the spectrum of a star with given spectral type interpolating the pickles database.
    The spectrum is normalized to unit mean.
    It assumes type V star.

    Inputs:
        wvs_or_filter_name: array of wavelenths in microns (or string with GPI band 'H', 'J', 'K1', 'K2', 'Y').
                (When using GPI spectral band wavelength samples are linearly spaced between the first and the last
                wavelength of the band.)
        star_type: 'A5','F4',... Is ignored if temperature is defined.
                If star_type is longer than 2 characters it is truncated.
        temperature: temperature of the star. Overwrite star_type if defined.

    Output:
        (wavelengths, spectrum) where
            wavelengths: Sampling in mum.
            spectrum: is the spectrum of the star for the given band.
    """

    if mute is None:
        mute = False

    if isinstance(wvs_or_filter_name, str):
        import pyklip.instruments.GPI as GPI
        sampling_wvs = GPI.get_gpi_wavelength_sampling(wvs_or_filter_name)
    else:
        sampling_wvs = wvs_or_filter_name

    sampling_wvs_unique = np.unique(sampling_wvs)


    if star_type is None:
        return sampling_wvs,None

    if len(star_type) > 2:
        star_type_selec = star_type[0:2]
    else:
        star_type_selec = star_type

    try:
        int(star_type_selec[1])
    except:
        try:
            star_type_selec = star_type[-3:-1]
            int(star_type_selec[1])
        except:
            if not mute:
                print("Returning None. Couldn't parse spectral type.")
            return sampling_wvs,None

    # Sory hard-coded type...
    if star_type_selec == "K8":
        star_type_selec = "K7"

    pykliproot = os.path.dirname(os.path.realpath(__file__))
    filename_temp_lookup = pykliproot+os.path.sep+"pickles"+os.path.sep+"mainseq_colors.txt"
    filename_pickles_lookup = pykliproot+os.path.sep+"pickles"+os.path.sep+"AA_README"

    #a = np.genfromtxt(filename_temp_lookup, names=True, delimiter=' ', dtype=None)

    # The interpolation is based on the temperature of the star
    # If the input was not the temperature then it is taken from the mainseq_colors.txt based on the input spectral type
    if temperature is None:
        #Read pickles list
        dict_temp = dict()
        with open(filename_temp_lookup, 'r') as f:
            for line in f:
                if line.startswith('#'):
                    pass
                else:
                    splitted_line = line.split()
                    # splitted_line[0]: spectral type F5 G0...
                    # splitted_line[2]: Temperature in K
                    dict_temp[splitted_line[0]] = splitted_line[2]

        try:
            target_temp = float(dict_temp[star_type_selec])
        except:
            if not mute:
                print("Returning None. Couldn't find a temperature for this spectral type in pickles mainseq_colors.txt.")
            return sampling_wvs,None
    else:
        target_temp = temperature

    # "AA_README" contains the list of the temperature for which a spectrum is available
    # Read it here
    dict_filename = dict()
    temp_list=[]
    with open(filename_pickles_lookup, 'r') as f:
        for line in f:
            if line.startswith('pickles_uk_'):
                splitted_line = line.split()
                # splitted_line[0]: Filename
                # splitted_line[1]: spectral type F5V G0III...
                # splitted_line[2]: Temperature in K

                #Check that the last character is numeric
                spec_type = splitted_line[1]
                if splitted_line[0][len(splitted_line[0])-1].isdigit() and not (spec_type.endswith('IV') or spec_type.endswith('I')):
                    dict_filename[float(splitted_line[2])] = splitted_line[0]
                    temp_list.append(float(splitted_line[2]))

    #temp_list = np.array(dict_filename.keys())
    temp_list = np.array(temp_list)
    # won't work for the hottest and coldest spectra.
    upper_temp, upper_temp_id = find_upper_nearest(temp_list,target_temp)
    lower_temp, lower_temp_id = find_lower_nearest(temp_list,target_temp)
    #print( upper_temp, upper_temp_id,lower_temp, lower_temp_id)

    upper_filename = dict_filename[upper_temp]
    lower_filename = dict_filename[lower_temp]

    upper_filename = pykliproot+os.path.sep+"pickles"+os.path.sep+upper_filename+".fits"
    lower_filename = pykliproot+os.path.sep+"pickles"+os.path.sep+lower_filename+".fits"


    hdulist = pyfits.open(upper_filename)
    cube = hdulist[1].data
    upper_wave = []
    upper_spec = []
    for wave_value,spec_value in cube:
        upper_wave.append(wave_value) # in angstrom
        upper_spec.append(spec_value)
    delta_wave = upper_wave[1]-upper_wave[0]
    upper_wave = np.array(upper_wave)/10**4 # in mum
    # upper_spec is a density spectrum in flux.A-1 so we need to multiply by delta_wave to integrate and get a flux.
    upper_spec = np.array(upper_spec)*delta_wave

    hdulist = pyfits.open(lower_filename)
    cube = hdulist[1].data
    lower_wave = []
    lower_spec = []
    for wave_value,spec_value in cube:
        lower_wave.append(wave_value) # in angstrom
        lower_spec.append(spec_value)
    lower_wave = np.array(lower_wave)/10**4 # in mum
    # lower_spec is a density spectrum in flux.A-1 so we need to multiply by delta_wave to integrate and get a flux.
    lower_spec = np.array(lower_spec)*delta_wave

    sampling_wvs_unique0 = np.insert(sampling_wvs_unique[:-1],0,sampling_wvs_unique[0])
    sampling_wvs_unique1 = np.insert(sampling_wvs_unique[1::],-1,sampling_wvs_unique[-1])
    upper_spec_unique = np.array([np.mean(upper_spec[np.where((upper_wave>wv0)*(upper_wave<wv1))]) for wv0,wv1 in zip(sampling_wvs_unique0,sampling_wvs_unique1)])
    lower_spec_unique = np.array([np.mean(lower_spec[np.where((lower_wave>wv0)*(lower_wave<wv1))]) for wv0,wv1 in zip(sampling_wvs_unique0,sampling_wvs_unique1)])

    # Sometimes the wavelength sampling is weird and the strategy above yields nans in the spectra.
    # When this happens we don't average out the spectra and takes the nearest available sample
    for k in range(np.size(upper_spec_unique)):
        if np.isnan(upper_spec_unique[k]):
            upper_spec_unique[k]= upper_spec[find_nearest(upper_wave,sampling_wvs_unique[k])[1]]
    for k in range(np.size(lower_spec_unique)):
        if np.isnan(lower_spec_unique[k]):
            lower_spec_unique[k]= lower_spec[find_nearest(lower_wave,sampling_wvs_unique[k])[1]]

    spec_pip_unique = ((target_temp-lower_temp)*upper_spec_unique+(upper_temp-target_temp)*lower_spec_unique)/(upper_temp-lower_temp)

    f = interp1d(sampling_wvs_unique, spec_pip_unique)
    spec_pip = f(sampling_wvs)

    return (sampling_wvs,spec_pip/np.nanmean(spec_pip))

def get_planet_spectrum(spectrum,wavelength,ori_wvs=None):
    """
    Get the normalized spectrum of a planet for a GPI spectral band or any wavelengths array.
    Spectra are extraced from .flx files from Mark Marley et al's models.

    Args:
        spectrum: Path of the .flx file containing the spectrum.
        wavelength: array of wavelenths in microns (or string with GPI band 'H', 'J', 'K1', 'K2', 'Y').
                (When using GPI spectral band wavelength samples are linearly spaced between the first and the last
                wavelength of the band.)

    Return:
        wavelengths: is the gpi sampling of the considered band in micrometer.
        spectrum: is the spectrum of the planet for the given band or wavelength array and normalized to unit mean.
    """


    if isinstance(spectrum, str):
        spec_data = []
        with open(spectrum, 'r') as f:
            for line in f:
                splitted_line = line.split()
                # splitted_line[0]: index
                # splitted_line[1]: wavelength (mum)
                # splitted_line[2]: T_brt
                # splitted_line[2]: flux in units of erg cm-2 sec-1 Hz-1 at the top of the planet's atmosphere

                try:
                    spec_data.append([float(splitted_line[0]),float(splitted_line[1]),float(splitted_line[2]),float(splitted_line[3])])
                except:
                    break

        spec_data = np.array(spec_data)
        N_samp = spec_data.shape[0]
        wave = spec_data[:,1]
        spec = spec_data[:,3]

    # Interpolate the spectrum on GPI sampling and convert F_nu to F_lambda
        spec = spec/wave**2
    else:
        wave = ori_wvs
        spec = spectrum


    # todo: check that it matches the actual sampling
    if isinstance(wavelength, str):
        import pyklip.instruments.GPI as GPI
        sampling_pip = GPI.get_gpi_wavelength_sampling(wavelength)
    else:
        sampling_pip = wavelength

    f = interp1d(wave, spec)
    # Interpolate the spectrum on GPI sampling and convert F_nu to F_lambda
    spec_pip = f(sampling_pip)/(sampling_pip**2)

    if 0:
        import matplotlib.pyplot as plt
        print((sampling_pip,spec_pip/np.nanmean(spec_pip)))
        plt.figure(2)
        wave_range = np.where((wave<sampling_pip[-1]) & (wave>sampling_pip[0]))
        plt.plot(wave[wave_range],spec[wave_range]/np.nanmean(spec[wave_range]),'r')
        plt.plot(sampling_pip,spec_pip/np.nanmean(spec_pip),'b.')
        plt.show()

    return (sampling_pip,spec_pip/np.nanmean(spec_pip))





def LSQ_scale_model_PSF(PSF_template,planet_image,a):
    return np.nansum((planet_image-a*PSF_template)**2,axis = (0,1))#/y_model

def extract_planet_spectrum(cube_para, position, PSF_cube_para, method = None,filter = None, mute = True):

    if isinstance(cube_para, basestring):
        hdulist = pyfits.open(cube_para)
        cube = hdulist[1].data
        exthdr = hdulist[1].header
        prihdr = hdulist[0].header
        hdulist.close()

        try:
            filter = prihdr['IFSFILT'].split('_')[1]
        except:
            if not mute:
                print("Couldn't find IFSFILT keyword in headers.")

    else:
        cube = cube_para

    if isinstance(PSF_cube_para, basestring):
        hdulist = pyfits.open(PSF_cube_para)
        PSF_cube = hdulist[1].data
        exthdr = hdulist[1].header
        prihdr = hdulist[0].header
        hdulist.close()

        sat_spot_spec = np.nanmax(PSF_cube,axis=(1,2))
        for l_id in range(PSF_cube.shape[0]):
            PSF_cube[l_id,:,:] /= sat_spot_spec[l_id]
    else:
        PSF_cube = PSF_cube_para

    row_cen,col_cen = extract_planet_centroid(cube, position, PSF_cube)

    nl,ny,nx = cube.shape
    row_id = np.round(row_cen)
    col_id = np.round(col_cen)
    #row_id,col_id = position
    nl_PSF,ny_PSF,nx_PSF = PSF_cube.shape

    row_m = np.floor(ny_PSF/2.0)
    row_p = np.ceil(ny_PSF/2.0)
    col_m = np.floor(nx_PSF/2.0)
    col_p = np.ceil(nx_PSF/2.0)
    cube_stamp = cube[:,np.max([0,(row_id-row_m)]):np.min([ny-1,(row_id+row_p)]), np.max([0,(col_id-col_m)]):np.min([nx-1,(col_id+col_p)])]
    nl_stamp, ny_stamp,nx_stamp = cube_stamp.shape


    stamp_x_grid, stamp_y_grid = np.meshgrid(np.arange(0,nx_stamp,1),np.arange(0,ny_stamp,1))
    r_stamp = np.sqrt((stamp_x_grid-(col_cen-(col_id-col_m)))**2 +(stamp_y_grid-(row_cen-(row_id-row_m)))**2)
    #stamp_mask = np.ones((stamp_nrow,stamp_ncol))
    #stamp_mask[np.where(r_stamp < 4.0)] = np.nan
    stamp_mask_small = np.ones((ny_stamp,nx_stamp))
    stamp_mask_small[np.where(r_stamp > 2.)] = np.nan
    stamp_cube_small_mask = np.tile(stamp_mask_small[None,:,:],(nl,1,1))


    # if 0:
        # plt.figure(6)
        # plt.subplot(2,2,1)
        # plt.imshow(stamp_mask_small,interpolation="nearest")
        # plt.subplot(2,2,2)
        # plt.imshow(cube_stamp[10,:,:]*stamp_mask_small,interpolation="nearest")
        # plt.show()

    if method is None or method == "max":
        spectrum = np.nanmax(cube_stamp*stamp_cube_small_mask,axis=(1,2))
    elif method == "aperture":
        spectrum = np.nansum(cube_stamp*stamp_cube_small_mask,axis=(1,2))
    elif method == "fit":
        spectrum = np.zeros((nl,))
        for k in range(nl):
            PSF_cube_slice = PSF_cube[k,:,:]/np.nanmax(PSF_cube[k,:,:])
            cube_stamp_slice = cube_stamp[k,:,:]
            param0 = np.nanmax(cube_stamp_slice)
            LSQ_func = lambda para: LSQ_scale_model_PSF(PSF_cube_slice,cube_stamp_slice,para)
            spectrum_fit[k] = minimize(LSQ_func,param0).x #, method="Nelder-Mead"
            # if 0:
                # plt.figure(1)
                # plt.subplot(2,2,1)
                # plt.imshow(cube_stamp_slice,interpolation="nearest")
                # plt.subplot(2,2,2)
                # plt.imshow(cube_stamp_slice*stamp_mask_small,interpolation="nearest")
                # plt.show()

    if 0:
        print(spectrum)
        # plt.figure(3)
        # plt.plot(get_gpi_wavelength_sampling(filter), spectrum)
        # plt.show()

    return get_gpi_wavelength_sampling(filter), spectrum



import os, subprocess
import astropy.io.fits as fits
from astropy import wcs
import numpy as np
import scipy.ndimage as ndimage
from scipy.ndimage.filters import median_filter

from pyklip.instruments.Instrument import Data
import pyklip.klip as klip

class Ifs(Data):
    """
    A spectral cube of Osiris IFS Data.

    Args:
        data_cube: FITS file list with 3D-cubes (Nwvs, Ny, Nx) with an osiris IFS data
        telluric_cube: single telluric reference FITS file with a 3D-cube (Nwvs, Ny, Nx) with an osiris IFS data.
        psf_cube_size: size of the psf cube to save (length along 1 dimension)
        coaddslices: if not None, combine (addition) slices together to reduce the size of the spectral cube.
                coaddslices should be an integer corresponding to the number of slices to be combined.

    Attributes:
        input: Array of shape (N,y,x) for N images of shape (y,x)
        centers: Array of shape (N,2) for N centers in the format [x_cent, y_cent]
        filenums: Array of size N for the numerical index to map data to file that was passed in
        filenames: Array of size N for the actual filepath of the file that corresponds to the data
        PAs: Array of N for the parallactic angle rotation of the target (used for ADI) [in degrees]
        wvs: Array of N wavelengths of the images (used for SDI) [in microns]. For polarization data, defaults to "None"
        IWA: a floating point scalar (not array). Specifies to inner working angle in pixels
        output: Array of shape (b, len(files), len(uniq_wvs), y, x) where b is the number of different KL basis cutoffs
        psfs: Spectral cube of size (Nwv, psfy, psfx) where psf_cube_size defines the size of psfy, psfx.
        psf_center: [x, y] location of the center of the PSF for a frame in self.psfs 
        flipx: True by default. Determines whether a relfection about the x axis is necessary to rotate image North-up East left
        nfiles: number of datacubes
        nwvs: number of wavelengths

    """
    # class initialization

    # Coonstructor
    def __init__(self, data_cube_list, telluric_cube,
                 guess_center=None,recalculate_center_cadi=False, centers = None,
                 psf_cube_size=21,
                 coaddslices=None, nan_mask_boxsize=0,median_filter_boxsize = 0,badpix2nan=False,ignore_PAs=True):
        super(Ifs, self).__init__()

        self.nfiles = len(data_cube_list)

        # read in the data
        self.filenums = []
        self.filenames = []
        self.prihdrs = []
        self.wvs = []
        self.centers = []
        self.OBFMXIM_list = []
        self.OBFMYIM_list = []
        self.PAs = []
        for k,data_cube in enumerate(data_cube_list):
            with fits.open(data_cube) as hdulist:
                print("Reading "+data_cube)
                tmp_input = np.rollaxis(np.rollaxis(hdulist[0].data,2),2,1)
                self.nwvs = tmp_input.shape[0]
                try:
                    self.input = np.concatenate((self.input,tmp_input),axis=0) # 3D cube, Nwvs, Ny, Nx
                except:
                    self.input = tmp_input
                self.prihdrs.append(hdulist[0].header)
                # Move dimensions of input array to match pyklip conventions
                self.filenums.extend(np.ones(self.nwvs)*k)
                self.filenames.extend([os.path.basename(data_cube),]*self.nwvs)
                # centers are at dim/2
                init_wv = self.prihdrs[k]["CRVAL1"]/1000. # wv for first slice in mum
                dwv = self.prihdrs[k]["CDELT1"]/1000. # wv interval between 2 slices in mum
                self.wvs.extend(np.arange(init_wv,init_wv+dwv*self.nwvs,dwv))

                # Plate scale of the spectrograph
                if "0.02" in self.prihdrs[k]["SSCALE"]:
                    self.platescale = 0.0203
                elif "0.035" in self.prihdrs[k]["SSCALE"]:
                    self.platescale = 0.0350
                elif "0.05" in self.prihdrs[k]["SSCALE"]:
                    self.platescale = 0.0500
                elif "0.1" in self.prihdrs[k]["SSCALE"]:
                    self.platescale = 0.1009


                self.PAs.extend(np.array([self.prihdrs[k]["PARANG"],]*self.nwvs))
                self.OBFMXIM_list.extend(np.array([self.prihdrs[k]["OBFMXIM"],]*self.nwvs))
                self.OBFMYIM_list.extend(np.array([self.prihdrs[k]["OBFMYIM"],]*self.nwvs))

                if guess_center is None:
                    self.centers.extend(np.array([[img.shape[1]/2., img.shape[0]/2.] for img in tmp_input]))
                else:
                    self.centers.extend(np.array([guess_center,]*self.nwvs))

        if centers is not None:
            self.centers = []
            for x,y in centers:
                self.centers.extend([[x,y],]*self.nwvs)

        self.wvs = np.array(self.wvs)
        self.centers = np.array(self.centers)

        # # TODO set the PAs right?
        if ignore_PAs:
            self.PAs = np.zeros(self.wvs.shape)

        if badpix2nan:
            box_w = 3
            smooth_input = median_filter(self.input,size=(box_w,box_w,box_w))
            res_input = np.abs((self.input - smooth_input))
            res_input = res_input/np.nanstd(res_input,axis=(1,2))[:,None,None]
            where_bad = np.where(res_input>5)
            self.input[where_bad] = np.nan

        self.input[np.where(self.input==0)] = np.nan
        # import matplotlib.pyplot as plt
        # for k in range(res_input.shape[0]):
        #     # plt.figure(1)
        #     # plt.imshow(res_input[k,::-1,:],interpolation="nearest")
        #     # plt.colorbar()
        #     plt.figure(2)
        #     plt.imshow(self.input[k,::-1,:],interpolation="nearest")
        #     plt.colorbar()
        #     # plt.figure(3)
        #     # res_input[k,::-1,:][np.where(res_input[k,::-1,:]>5)] = np.nan
        #     # plt.imshow(res_input[k,::-1,:],interpolation="nearest")
        #     # plt.colorbar()
        #     plt.show()

        # import matplotlib.pyplot as plt
        # for k in range(self.nwvs):
        #     plt.imshow(self.input[50*k,::-1,:])
        #     plt.colorbar()
        #     plt.show()
        # read in the psf cube
        with fits.open(telluric_cube) as hdulist:
            psfs = hdulist[0].data # Nwvs, Ny, Nx
            # Move dimensions of input array to match pyklip conventions
            psfs = np.rollaxis(np.rollaxis(psfs,2),2,1)

            # The definition of psfs_wvs requires that no wavelengths has been skipped in the input files
            # But it works with keepslices
            self.psfs_wvs = np.arange(init_wv,init_wv+dwv*self.nwvs,dwv)

            # trim the cube
            pixelsbefore = psf_cube_size//2
            pixelsafter = psf_cube_size - pixelsbefore

            psfs = np.pad(psfs,((0,0),(pixelsbefore,pixelsafter),(pixelsbefore,pixelsafter)),mode="constant",constant_values=0)
            psfs_centers = np.array([np.unravel_index(np.nanargmax(img),img.shape) for img in psfs])
            # Change center index order to match y,x convention
            psfs_centers = [(cent[1],cent[0]) for cent in psfs_centers]
            psfs_centers = np.array(psfs_centers)
            center0 = np.median(psfs_centers,axis=0)

            # TODO Calculate precise centroid
            from pyklip.fakes import gaussfit2d
            psfs_centers = []
            self.star_peaks = []
            self.psfs = np.zeros((psfs.shape[0],psf_cube_size,psf_cube_size))
            for k,im in enumerate(psfs):
                corrflux, fwhm, spotx, spoty = gaussfit2d(im, center0[0], center0[1], searchrad=5, guessfwhm=3, guesspeak=np.nanmax(im), refinefit=True)
                #spotx, spoty = center0
                psfs_centers.append((spotx, spoty))
                self.star_peaks.append(corrflux)

                # Get the closest pixel
                xarr_spot = int(np.round(spotx))
                yarr_spot = int(np.round(spoty))
                # Extract a stamp around the sat spot
                stamp = im[(yarr_spot-pixelsbefore):(yarr_spot+pixelsafter),\
                                (xarr_spot-pixelsbefore):(xarr_spot+pixelsafter)]
                # Define coordinates grids for the stamp
                stamp_x, stamp_y = np.meshgrid(np.arange(psf_cube_size, dtype=np.float32),
                                               np.arange(psf_cube_size, dtype=np.float32))
                # Calculate the shift of the sat spot centroid relative to the closest pixel.
                dx = spotx-xarr_spot
                dy = spoty-yarr_spot

                # The goal of the following section is to remove the local background (or sky) around the sat spot.
                # The plane is defined by 3 constants (a,b,c) such that z = a*x+b*y+c
                # In order to do so we fit a 2D plane to the stamp after having masked the sat spot (centered disk)
                stamp_r = np.sqrt((stamp_x-dx-psf_cube_size//2)**2+(stamp_y-dy-psf_cube_size//2)**2)
                from copy import copy
                stamp_masked = copy(stamp)
                stamp_x_masked = stamp_x-dx
                stamp_y_masked = stamp_y-dy
                stamp_center = np.where(stamp_r<7)
                stamp_masked[stamp_center] = np.nan
                stamp_x_masked[stamp_center] = np.nan
                stamp_y_masked[stamp_center] = np.nan
                background_med =  np.nanmedian(stamp_masked)
                stamp_masked = stamp_masked - background_med
                #Solve 2d linear fit to remove background
                xx = np.nansum(stamp_x_masked**2)
                yy = np.nansum(stamp_y_masked**2)
                xy = np.nansum(stamp_y_masked*stamp_x_masked)
                xz = np.nansum(stamp_masked*stamp_x_masked)
                yz = np.nansum(stamp_y_masked*stamp_masked)
                #Cramer's rule
                a = (xz*yy-yz*xy)/(xx*yy-xy*xy)
                b = (xx*yz-xy*xz)/(xx*yy-xy*xy)
                stamp = stamp - (a*(stamp_x-dx)+b*(stamp_y-dy) + background_med)

                stamp = ndimage.map_coordinates(stamp, [stamp_y+dy, stamp_x+dx])

                self.psfs[k,:,:] = stamp

            # import matplotlib.pyplot as plt
            # plt.figure(1)
            # plt.imshow(np.nanmean(psfs,axis=0))
            # plt.figure(2)
            # plt.imshow(np.nanmean(self.psfs,axis=0))
            # plt.show()

            # Spectrum of the telluric star
            self.star_peaks = np.array(self.star_peaks)
            # TODO include brightness of the telluric star
            self.dn_per_contrast = np.array([self.star_peaks[np.where(self.psfs_wvs==wv)[0]] for wv in self.wvs])

        # we don't need to flip x for North Up East left
        self.flipx = False

        # I have no idea
        self.IWA = 0.0
        # Infinity...
        self.OWA = 10000


        self._output = None

        if coaddslices is not None:

            # N_chunks = self.psfs.shape[0]//coaddslices
            N_chunks = self.nwvs//coaddslices
            self.psfs = np.array([np.nanmean(self.psfs[k*coaddslices:(k+1)*coaddslices,:,:],axis=0) for k in range(N_chunks)])
            self.psfs_wvs = np.array([np.nanmean(self.psfs_wvs[k*coaddslices:(k+1)*coaddslices]) for k in range(N_chunks)])

            new_wvs = []
            new_filenums = []
            new_centers = []
            new_PAs = []
            new_OBFMXIM = []
            new_OBFMYIM = []
            new_filenames = []
            new_dn_per_contrast = []
            # new_wcs = []
            for k in range(self.nfiles):
                tmp_input = copy(self.input[k*self.nwvs:(k+1)*self.nwvs,:,:])
                tmp_input = np.array([np.nanmean(tmp_input[l*coaddslices:(l+1)*coaddslices,:,:],axis=0) for l in range(N_chunks)])
                try:
                    new_input = np.concatenate((new_input,tmp_input),axis=0) # 3D cube, Nwvs, Ny, Nx
                except:
                    new_input = tmp_input
                new_nwvs = tmp_input.shape[0]

                new_wvs.extend(np.array([np.nanmean(self.wvs[k*self.nwvs:(k+1)*self.nwvs][l*coaddslices:(l+1)*coaddslices]) for l in range(N_chunks)]))
                new_filenums.extend(np.array([self.filenums[k*self.nwvs:(k+1)*self.nwvs][l*coaddslices] for l in range(N_chunks)]))
                new_centers.extend([np.nanmean(self.centers[k*self.nwvs:(k+1)*self.nwvs,:][l*coaddslices:(l+1)*coaddslices,:],axis=0) for l in range(N_chunks)])
                new_PAs.extend([self.PAs[k*self.nwvs:(k+1)*self.nwvs][l*coaddslices] for l in range(N_chunks)])
                new_OBFMXIM.extend([self.OBFMXIM_list[k*self.nwvs:(k+1)*self.nwvs][l*coaddslices] for l in range(N_chunks)])
                new_OBFMYIM.extend([self.OBFMYIM_list[k*self.nwvs:(k+1)*self.nwvs][l*coaddslices] for l in range(N_chunks)])
                new_filenames.extend([self.filenames[k*self.nwvs:(k+1)*self.nwvs][l*coaddslices] for l in range(N_chunks)])
                new_dn_per_contrast.extend([np.nanmean(self.dn_per_contrast[k*self.nwvs:(k+1)*self.nwvs][l*coaddslices:(l+1)*coaddslices]) for l in range(N_chunks)])
                # new_wcs.extend([self.wcs[k*self.nwvs:(k+1)*self.nwvs][l*coaddslices] for l in range(N_chunks)])


            self.input = new_input
            self.wvs = np.array(new_wvs)
            self.nwvs = new_nwvs
            self.filenums = np.array(new_filenums)
            self.centers = np.array(new_centers)
            self.PAs = np.array(new_PAs)
            self.OBFMXIM_list = np.array(new_OBFMXIM)
            self.OBFMYIM_list = np.array(new_OBFMYIM)
            self.filenames = new_filenames
            self.dn_per_contrast = np.array(new_dn_per_contrast)
            # self.wcs = new_wcs
        self.dn_per_contrast = np.squeeze(self.dn_per_contrast)

        # # TODO: need to check how it works (cf GPI)
        # self.wcs = np.array([None for _ in range(self.nfiles * self.nwvs)])
        # Creating WCS info for OSIRIS
        self.wcs = []
        for vert_angle in self.PAs:
            w = wcs.WCS()
            vert_angle = np.radians(vert_angle)
            pc = np.array([[(-1)*np.cos(vert_angle), (-1)*-np.sin(vert_angle)],[np.sin(vert_angle), np.cos(vert_angle)]])
            cdmatrix = pc * self.platescale /3600.
            w.wcs.cd = cdmatrix
            self.wcs.append(w)
        self.wcs = np.array(self.wcs)

        # import matplotlib.pyplot as plt
        # for k in range(self.nwvs):
        #     plt.imshow(self.input[k,::-1,:])
        #     plt.colorbar()
        #     plt.show()

        if median_filter_boxsize != 0:
            self.input = median_filter(self.input,size=(1,median_filter_boxsize,median_filter_boxsize))
            self.psfs = median_filter(self.psfs,size=(1,median_filter_boxsize,median_filter_boxsize))

        if nan_mask_boxsize != 0:
            # zeros are nans, and anything adjacient to a pixel less than zero is 0.
            input_nans = np.where(np.isnan(self.input))
            self.input[input_nans] = 0
            input_minfilter = ndimage.minimum_filter(self.input, (0, nan_mask_boxsize, nan_mask_boxsize))
            self.input[np.where(input_minfilter <= 0)] = np.nan
            self.input[:,0:nan_mask_boxsize//2,:] = np.nan
            self.input[:,-nan_mask_boxsize//2+1::,:] = np.nan
            self.input[:,:,0:nan_mask_boxsize//2] = np.nan
            self.input[:,:,-nan_mask_boxsize//2+1::] = np.nan

        # for wv_index in range(self.psfs.shape[0]):
        #     model_psf = self.psfs[wv_index, :, :]
        #     import matplotlib.pyplot as plt
        #     plt.imshow(model_psf)
        #     plt.show()
        # import matplotlib.pyplot as plt
        # for k in range(self.nwvs):
        #     plt.imshow(self.input[10*k,::-1,:])
        #     plt.colorbar()
        #     plt.show()

        # Required for automatically querying Simbad for the spectral type of the star.
        self.object_name = "HR8799"#self.prihdr["OBJECT"]

        if recalculate_center_cadi:
            for k in range(self.nfiles):
                tmp_input = copy(self.input[k*self.nwvs:(k+1)*self.nwvs,:,:])
                if guess_center is None:
                    xcen0,ycen0 = tmp_input.shape[2]/2., tmp_input.shape[1]/2.
                else:
                    xcen0,ycen0 = guess_center

                range_list = [100,20,4,1]
                samples = 10
                for it,width in enumerate(range_list):
                    x_list,y_list = np.linspace(xcen0-width/2.,xcen0+width/2.,samples),np.linspace(ycen0-width/2.,ycen0+width/2.,samples)
                    # print(x_list,y_list)
                    xcen_grid,ycen_grid = np.meshgrid(x_list,y_list)
                    cost_func  = np.zeros(xcen_grid.shape)
                    cost_func.shape = [np.size(cost_func)]

                    import multiprocessing as mp
                    import itertools
                    self.N_threads = mp.cpu_count()
                    pool = mp.Pool(processes=self.N_threads)
                    #multitask this
                    outputs_list = pool.map(casdi_residual_star, zip(xcen_grid.ravel(),
                                                                                ycen_grid.ravel(),
                                                                                itertools.repeat(tmp_input),
                                                                                itertools.repeat(self.wvs[k*self.nwvs:(k+1)*self.nwvs])))

                    for l,out in enumerate(outputs_list):
                        cost_func[l] = out
                    pool.close()

                    xcen0 = xcen_grid.ravel()[np.argmin(cost_func)]
                    ycen0 = ycen_grid.ravel()[np.argmin(cost_func)]

                    # import matplotlib.pyplot as plt
                    # cost_func.shape = (samples,samples)
                    # plt.figure(1)
                    # plt.subplot(2,1,1)
                    # plt.imshow(xcen_grid[::-1,:],interpolation="nearest")
                    # plt.colorbar()
                    # plt.subplot(2,1,2)
                    # plt.imshow(ycen_grid[::-1,:],interpolation="nearest")
                    # plt.colorbar()
                    # plt.figure(2)
                    # plt.imshow(cost_func[::-1,:],interpolation="nearest")
                    # plt.figure(3)
                    # plt.plot(self.wvs[k*self.nwvs:(k+1)*self.nwvs])
                    # plt.show()
                    # print(k,xcen0,ycen0)
                self.centers[k*self.nwvs:(k+1)*self.nwvs,:] = np.array([(xcen0,ycen0),]*tmp_input.shape[0])


        # else:
        #     # from scipy.optimize import leastsq
        #     # LSQ_func = lambda para: casdi_residual(para[0],para[1],self.input,self.wvs,nan2zero=True)
        #     # new_cent = leastsq(LSQ_func,(img.shape[1]/2.-sep_planet/ 0.02, img.shape[0]/2.))
        #     from scipy.optimize import minimize
        #     LSQ_func = lambda para: np.nanvar(casdi_residual(para[0],para[1],self.input,self.wvs,nan2zero=False))
        #     new_cent = minimize(LSQ_func,(img.shape[1]/2.-sep_planet/ 0.02, img.shape[0]/2.),method="nelder-mead",options={'disp':True})
        #     # casdi_residual(cent[0],cent[1],self.input,self.wvs)
        #     print("old",(img.shape[1]/2.-sep_planet/ 0.02, img.shape[0]/2.))
        #     print("new_cent",new_cent.x)
        #     exit()

    ################################
    ### Instance Required Fields ###
    ################################

    @property
    def input(self):
        return self._input
    @input.setter
    def input(self, newval):
        self._input = newval

    @property
    def centers(self):
        return self._centers
    @centers.setter
    def centers(self, newval):
        self._centers = newval

    @property
    def filenums(self):
        return self._filenums
    @filenums.setter
    def filenums(self, newval):
        self._filenums = newval

    @property
    def filenames(self):
        return self._filenames
    @filenames.setter
    def filenames(self, newval):
        self._filenames = newval

    @property
    def PAs(self):
        return self._PAs
    @PAs.setter
    def PAs(self, newval):
        self._PAs = newval

    @property
    def wvs(self):
        return self._wvs
    @wvs.setter
    def wvs(self, newval):
        self._wvs = newval

    @property
    def wcs(self):
        return self._wcs
    @wcs.setter
    def wcs(self, newval):
        self._wcs = newval


    @property
    def IWA(self):
        return self._IWA
    @IWA.setter
    def IWA(self, newval):
        self._IWA = newval

    @property
    def output(self):
        return self._output
    @output.setter
    def output(self, newval):
        self._output = newval


    ###############
    ### Methods ###
    ###############

    def readdata(self, filepaths):
        """
        Reads in the data from the files in the filelist and writes them to fields
        """
        pass


    def savedata(self, filepath, data,center=None, klipparams=None, filetype="", zaxis=None , more_keywords=None):
        """
        Save OSIRIS Data.

        Note: In principle, the function only works inside klip_dataset(). In order to use it outside of klip_dataset,
            you need to define the following attribute:
                dataset.output_centers = dataset.centers

        Args:
            filepath: path to file to output
            data: 2D or 3D data to save
            center: center of the image to be saved in the header as the keywords PSFCENTX and PSFCENTY in pixels.
                The first pixel has coordinates (0,0)
            klipparams: a string of klip parameters
            filetype: filetype of the object (e.g. "KL Mode Cube", "PSF Subtracted Spectral Cube")
            zaxis: a list of values for the zaxis of the datacub (for KL mode cubes currently)
            more_keywords (dictionary) : a dictionary {key: value, key:value} of header keywords and values which will
                                         written into the primary header
        
        """
        hdulist = fits.HDUList()
        hdulist.append(fits.PrimaryHDU(data=data,header=self.prihdrs[0]))

        # save all the files we used in the reduction
        # we'll assume you used all the input files
        # remove duplicates from list
        filenames = np.unique(self.filenames)
        nfiles = np.size(filenames)
        hdulist[0].header["DRPNFILE"] = (nfiles, "Num raw files used in pyKLIP")
        for i, filename in enumerate(filenames):
            hdulist[0].header["FILE_{0}".format(i)] = filename + '.fits'

        # write out psf subtraction parameters
        # get pyKLIP revision number
        pykliproot = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        # the universal_newline argument is just so python3 returns a string instead of bytes
        # this will probably come to bite me later
        try:
            pyklipver = subprocess.check_output(['git', 'rev-parse', '--short', 'HEAD'], cwd=pykliproot, universal_newlines=True).strip()
        except:
            pyklipver = "unknown"
        hdulist[0].header['PSFSUB'] = ("pyKLIP", "PSF Subtraction Algo")
        hdulist[0].header.add_history("Reduced with pyKLIP using commit {0}".format(pyklipver))
        hdulist[0].header['CREATOR'] = "pyKLIP-{0}".format(pyklipver)

        # store commit number for pyklip
        hdulist[0].header['pyklipv'] = (pyklipver, "pyKLIP version that was used")

        if klipparams is not None:
            hdulist[0].header['PSFPARAM'] = (klipparams, "KLIP parameters")
            hdulist[0].header.add_history("pyKLIP reduction with parameters {0}".format(klipparams))

        # write z axis units if necessary
        if zaxis is not None:
            # Writing a KL mode Cube
            if "KL Mode" in filetype:
                hdulist[0].header['CTYPE3'] = 'KLMODES'
                # write them individually
                for i, klmode in enumerate(zaxis):
                    hdulist[0].header['KLMODE{0}'.format(i)] = (klmode, "KL Mode of slice {0}".format(i))
                hdulist[0].header['CUNIT3'] = "N/A"
                hdulist[0].header['CRVAL3'] = 1
                hdulist[0].header['CRPIX3'] = 1.
                hdulist[0].header['CD3_3'] = 1.


        #use the dataset center if none was passed in
        center = self.output_centers[0]
        if center is not None:
            hdulist[0].header.update({'PSFCENTX': center[0], 'PSFCENTY': center[1]})
            hdulist[0].header.update({'CRPIX1': center[0], 'CRPIX2': center[1]})
            hdulist[0].header.add_history("Image recentered to {0}".format(str(center)))

        try:
            hdulist.writeto(filepath, overwrite=True)
        except TypeError:
            hdulist.writeto(filepath, clobber=True)
        hdulist.close()

    def calibrate_output(self, img, spectral=False, units="contrast"):
        """
       Calibrates the flux of an output image. Can either be a broadband image or a spectral cube depending
        on if the spectral flag is set.

        Assumes the broadband flux calibration is just multiplication by a single scalar number whereas spectral
        datacubes may have a separate calibration value for each wavelength

        Args:
            img: unclaibrated image.
                 If spectral is not set, this can either be a 2-D or 3-D broadband image
                 where the last two dimensions are [y,x]
                 If specetral is True, this is a 3-D spectral cube with shape [wv,y,x]
            spectral: if True, this is a spectral datacube. Otherwise, it is a broadband image.
            units: currently only support "contrast" w.r.t central star

        Return:
            img: calibrated image of the same shape (this is the same object as the input!!!)
        """
        if units == "contrast":
            if spectral:
                # spectral cube, each slice needs it's own calibration
                numwvs = img.shape[0]
                img /= self.dn_per_contrast[:numwvs, None, None]
            else:
                # broadband image
                img /= np.nanmean(self.dn_per_contrast)
            self.flux_units = "contrast"

        return img


def casdi_residual_star(params):
    """
    Convert `f([1,2])` to `f(1,2)` call.
    It allows one to call casdi_residual() with a tuple of parameters.
    """
    return np.nanvar(casdi_residual(*params))

def casdi_residual(xcen,ycen,input,wvs,nan2zero = False):
    input_scaled = np.zeros(input.shape)
    ref_wv = np.mean(wvs)

    for k,wv in enumerate(wvs):
        input_scaled[k,:,:] = klip.align_and_scale(input[k,:,:],(xcen,ycen),(xcen,ycen),ref_wv/wv)

    # input_sub = np.zeros(input.shape)
    # lib_size = np.max([np.size(wvs)/10,10])
    # for k,wv in enumerate(wvs):
    #     # print(k,np.size(wvs))
    #     input_sub[k,:,:] = input_scaled[k,:,:] - np.nanmedian(input_scaled[np.max([0,k-lib_size]):np.min([np.size(wvs),k+lib_size]),:,:],axis=0)

    import warnings
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        input_sub = input_scaled - np.nanmedian(input_scaled,axis=0)[None,:,:]

    if nan2zero:
        input_sub[np.where(np.isnan(input_sub))] = 0

    # for k,wv in enumerate(wvs):
    #     input_sub[k,:,:] = klip.align_and_scale(input_sub[k,:,:],(xcen,ycen),(xcen,ycen),wv/ref_wv)
    # print(xcen,ycen,np.nansum(input_sub**2),np.nanvar(input_sub.ravel()))
    # import matplotlib.pyplot as plt
    # plt.figure(1)
    # plt.imshow(np.nanmedian(input,axis=0))
    # print(np.nanvar(np.nanmedian(input,axis=0).ravel()))
    # plt.colorbar()
    # plt.figure(2)
    # plt.imshow(np.nanmedian(input_scaled,axis=0))
    # print(np.nanvar(np.nanmedian(input_scaled,axis=0).ravel()))
    # plt.colorbar()
    # plt.figure(3)
    # plt.imshow(np.nanmedian(input_sub,axis=0))
    # print(np.nanvar(np.nanmedian(input_sub,axis=0).ravel()))
    # plt.colorbar()
    # plt.show()

    return input_sub.ravel()

def determine_mosaic_offsets_from_header(prihdr_list):

    OBFMXIM_list = []
    OBFMYIM_list = []
    parang_list = []
    vd_InstAngl_list = []
    for k,prihdr in enumerate(prihdr_list):
        OBFMXIM_list.append(float(prihdr["OBFMXIM"]))
        OBFMYIM_list.append(float(prihdr["OBFMYIM"]))
        parang_list.append(float(prihdr["PARANG"]))
        vd_InstAngl_list.append(float(prihdr["INSTANGL"]))

    vd_C0 = OBFMXIM_list
    vd_C1 = OBFMYIM_list
    md_Coords = np.array([vd_C0,vd_C1])
    vd_InstAngl = np.array(vd_InstAngl_list)

    if "0.02" in prihdr["SSCALE"]:
        d_Scale = 0.0203
    elif "0.035" in prihdr["SSCALE"]:
        d_Scale = 0.0350
    elif "0.05" in prihdr["SSCALE"]:
        d_Scale = 0.0500
    elif "0.1" in prihdr["SSCALE"]:
        d_Scale = 0.1009
    else:
        d_Scale = 0.0203

    vd_CoordsNX =   (md_Coords[0,0] - md_Coords[0,:]) * (35.6 * (0.0397/d_Scale))
    vd_CoordsNY  =  (md_Coords[1,0] - md_Coords[1,:]) * (35.6 * (0.0397/d_Scale))

    vd_InstAngl = np.deg2rad(vd_InstAngl)
    md_Offsets = np.array([vd_CoordsNX * np.cos(vd_InstAngl) + vd_CoordsNY * np.sin(vd_InstAngl),
                   (-1.)*vd_CoordsNX * np.sin(vd_InstAngl) + vd_CoordsNY * np.cos(vd_InstAngl)])

    delta_x = -(md_Offsets[1,:]-md_Offsets[1,0])
    delta_y = -(md_Offsets[0,:]-md_Offsets[0,0])

    return delta_x,delta_y


def determine_mosaic_offsets_from_speckles(input_list):
    ncubes,nz,ny,nx = input_list.shape
    flat_inputs = np.nansum(input_list,axis=1)
    flat_inputs[np.where(flat_inputs==0)] = np.nan
    flat_inputs = flat_inputs/np.nanstd(flat_inputs,axis=(1,2))[:,None,None]
    print(flat_inputs.shape)
    print(ncubes,nz,ny,nx )
    # exit()

    # a = np.zeros((11,11))
    # b = np.zeros((11,11))
    # a[5:7,5:7] = 1
    # b[6:8,6:8] = 1
    #
    # from image_registration.image_registration import chi2_shift
    # xoff,yoff,exoff,eyoff = chi2_shift(a,a,np.ones((11,11)),return_error=True,upsample_factor='auto')
    # print("coucou")
    # print(xoff,yoff,exoff,eyoff)
    # exit()

    # plt.show()

    # dx_map = np.zeros((ncubes,ncubes))
    # dy_map = np.zeros((ncubes,ncubes))
    # edx_map = np.zeros((ncubes,ncubes))
    # edy_map = np.zeros((ncubes,ncubes))
    # noise = np.ones((ny,nx))
    # from image_registration.image_registration import chi2_shift
    # for k in np.arange(0,ncubes):
    #     for l in np.arange(0,ncubes):
    #         dx_map[k,l],dy_map[k,l],edx_map[k,l],edy_map[k,l] = chi2_shift(flat_inputs[k],flat_inputs[l],noise,return_error=True,upsample_factor='auto')
    #
    #
    # import matplotlib.pyplot as plt
    # plt.subplot(2,2,1)
    # plt.imshow(dx_map,interpolation="nearest")
    # plt.colorbar()
    # plt.subplot(2,2,2)
    # plt.imshow(dy_map,interpolation="nearest")
    # plt.colorbar()
    # plt.subplot(2,2,3)
    # plt.imshow(edx_map,interpolation="nearest")
    # plt.colorbar()
    # plt.subplot(2,2,4)
    # plt.imshow(edy_map,interpolation="nearest")
    # plt.colorbar()
    # plt.show()

    delta_x = [0]
    delta_y = [0]
    delta_x2 = [0]
    delta_y2 = [0]
    from image_registration.image_registration import chi2_shift
    from image_registration.image_registration.fft_tools import shift
    for k in np.arange(1,ncubes):
        im1 = flat_inputs[0]
        im2 = flat_inputs[k]
        noise = np.ones(im2.shape)
        xoff,yoff,exoff,eyoff = chi2_shift(im1,im2,noise/10.,return_error=True,upsample_factor='auto')
        # print("bonjour")
        # print(xoff,yoff,exoff,eyoff)
        # print(xoff-0.5,yoff-0.5,exoff,eyoff)
        # exit()
        delta_x.append(xoff)
        delta_y.append(yoff)
        im1 = flat_inputs[k-1]
        im2 = flat_inputs[k]
        noise = np.ones(im2.shape)
        xoff2,yoff2,exoff,eyoff = chi2_shift(im1,im2,noise,return_error=True,upsample_factor='auto')
        # print(xoff2+delta_x2[-1],yoff2+delta_y2[-1],xoff2,yoff2,exoff,eyoff)
        delta_x2.append(xoff2+delta_x2[-1])
        delta_y2.append(yoff2+delta_y2[-1])

        # import matplotlib.pyplot as plt
        # plt.subplot(2,3,1)
        # plt.imshow(flat_inputs[0],interpolation="nearest")
        # plt.colorbar()
        # plt.subplot(2,3,2)
        # plt.imshow(flat_inputs[k],interpolation="nearest")
        # plt.colorbar()
        # plt.subplot(2,3,3)
        # plt.imshow(flat_inputs[k-1],interpolation="nearest")
        # plt.colorbar()
        # plt.subplot(2,3,4)
        # res = flat_inputs[0]-shift.shiftnd(flat_inputs[k],(-yoff+0.5,-xoff+0.5))
        # print(np.sqrt(np.nansum(res**2)))
        # plt.imshow(res,interpolation="nearest")
        # plt.colorbar()
        # plt.subplot(2,3,6)
        # res = flat_inputs[k-1]-shift.shiftnd(flat_inputs[k],(-yoff2+0.5,-xoff2+0.5))
        # print(np.sqrt(np.nansum(res**2)))
        # plt.imshow(res,interpolation="nearest")
        # plt.colorbar()
        # plt.show()

    # import matplotlib.pyplot as plt
    # plt.figure(10)
    # for k in range(ncubes):
    #     plt.subplot(2,ncubes,k+1)
    #     plt.imshow(flat_inputs[k,::-1,:],interpolation="nearest")
    #     plt.subplot(2,ncubes,ncubes+k+1)
    #     plt.imshow(flat_inputs[0,::-1,:]-shift.shiftnd(flat_inputs[k,::-1,:],(-delta_y[k],-delta_x[k])),interpolation="nearest")

    # import matplotlib.pyplot as plt
    # plt.subplot(1,2,1)
    # plt.plot(delta_x,"r")
    # plt.plot(delta_x2,"b")
    # plt.subplot(1,2,2)
    # plt.plot(delta_y,"r")
    # plt.plot(delta_y2,"b")
    # plt.show()


    return delta_x,delta_y,delta_x2,delta_y2


__author__ = 'JB'


import numpy as np
from copy import copy
from  glob import glob
import csv
import os

def mask_known_objects(cube,fakeinfohdr,object_name,pix2as,center,MJDOBS=None,OI_list_folder=None,ignore_fakes = False,fakes_only = False,
                          include_speckles = False,IWA=None,OWA=None, mask_radius = 7):
    """

    Mask point sources in cube with an NaN aperture.

    Args:
        cube: Image or cube to be masked.
        fakeinfohdr: fits file header containing the injected planets related keywords.
        object_name: Name of the star being observed.
        pix2as: platescale.
        center: Center of the image.
        MJDOBS: Julian date of the observation. (needed when OI_list_folder is not None)
        OI_list_folder: List of Object of Interest (OI) that should be masked from any standard deviation
                        calculation. See the online documentation for instructions on how to define it.
        xy: Boolean. Returns the planets coordinate with x,y coordinates in pixels
        pa_sep: Boolean. Returns the planets coordinates as position angle (in arcsec), separation (in pix)
        ignore_fakes: Don't return fake planets.
        fakes_only: Returns only fake planets.
        include_speckles: Include speckles from the list of object of interest (OI_list_folder)
        IWA: Inner working angle (in pixels).
        OWA: Outer working angle (in pixels).
        mask_radius: Radius of the mask in pixels. (default = 7 pixels)

    Return: Masked cube.
    """

    row_vec,col_vec = get_pos_known_objects(fakeinfohdr,object_name,pix2as,center=center,MJDOBS=MJDOBS,OI_list_folder=OI_list_folder,
                          xy = False,pa_sep = False,ignore_fakes = ignore_fakes,fakes_only = fakes_only,
                          include_speckles = include_speckles,IWA=IWA,OWA=OWA)

    cube_cpy = copy(cube)

    if np.size(cube_cpy.shape) == 3:
        nl,ny,nx = cube_cpy.shape
    elif np.size(cube_cpy.shape) == 2:
        ny,nx = cube_cpy.shape
        cube_cpy = cube_cpy[None,:]
        nl = 1

    width = 2*mask_radius+1
    stamp_x_grid, stamp_y_grid = np.meshgrid(np.arange(0,width,1)-width/2,np.arange(0,width,1)-width/2)
    stamp_mask = np.ones((width,width))
    r_stamp = abs((stamp_x_grid) +(stamp_y_grid)*1j)
    stamp_mask[np.where(r_stamp < mask_radius)] = np.nan

    row_m = int(np.floor(width/2.0))
    row_p = int(np.ceil(width/2.0))
    col_m = int(np.floor(width/2.0))
    col_p = int(np.ceil(width/2.0))

    for row,col in zip(row_vec,col_vec):
        k = int(round(row))
        l = int(round(col))

        cube_cpy[:,(k-row_m):(k+row_p), (l-col_m):(l+col_p)] = np.tile(stamp_mask,(nl,1,1)) * cube_cpy[:,(k-row_m):(k+row_p), (l-col_m):(l+col_p)]

    return np.squeeze(cube_cpy)


def get_pos_known_objects(fakeinfohdr,object_name,pix2as,center=None,MJDOBS=None,OI_list_folder=None,xy = False,pa_sep = False,ignore_fakes = False,fakes_only = False,
                          include_speckles = False,IWA=None,OWA=None):
    """
    
    Return the position of real and/or simulated planets in an image based on its headers.

    Args:
        fakeinfohdr: fits file header containing the injected planets related keywords.
        object_name: Name of the star being observed.
        pix2as: platescale.
        center: Center of the image (not needed if pa_sep = True).
        MJDOBS: Julian date of the observation. (needed when OI_list_folder is not None)
        OI_list_folder: List of Object of Interest (OI) that should be masked from any standard deviation
                        calculation. See the online documentation for instructions on how to define it.
        xy: Boolean. Returns the planets coordinate with x,y coordinates in pixels
        pa_sep: Boolean. Returns the planets coordinates as position angle (in arcsec), separation (in pix)
        ignore_fakes: Don't return fake planets.
        fakes_only: Returns only fake planets.
        include_speckles: Include speckles from the list of object of interest (OI_list_folder)
        IWA: Inner working angle (in pixels).
        OWA: Outer working angle (in pixels).

    Return: (sep_vec,pa_vec) or (x_vec,y_vec) or (row_vec,col_vec) (default)
        Objects coordinates (real/fakes/others). For e.g., sep_vec,pa_vec are vectors and (sep_vec[0],pa_vec[0]) is the
        coordinate of the first object and so on...
    """
    x_vec = []
    y_vec = []
    col_vec = []
    row_vec = []
    pa_vec = []
    sep_vec = []


    if object_name is not None:
        _object_name = copy(object_name).replace(" ","_")

        if not fakes_only and MJDOBS is not None and OI_list_folder is not None:
            object_GOI_filename = OI_list_folder+os.path.sep+_object_name+'_GOI.csv'
            if len(glob(object_GOI_filename)) != 0:
                with open(object_GOI_filename, 'r') as csvfile_GOI_list:
                    GOI_list_reader = csv.reader(csvfile_GOI_list, delimiter=';')
                    GOI_csv_as_list = list(GOI_list_reader)
                    attrib_name = GOI_csv_as_list[0]
                    GOI_list = np.array(GOI_csv_as_list[1:len(GOI_csv_as_list)])

                    pa_id = attrib_name.index("PA")
                    sep_id = attrib_name.index("SEP")
                    MJDOBS_id = attrib_name.index("MJDOBS")
                    STATUS_id = attrib_name.index("STATUS")

                    MJDOBS_arr = np.array([ float(it) for it in GOI_list[:,MJDOBS_id]])
                    MJDOBS_unique = np.unique(MJDOBS_arr)
                    MJDOBS_closest_id = np.argmin(np.abs(MJDOBS_unique-MJDOBS))
                    MJDOBS_closest = MJDOBS_unique[MJDOBS_closest_id]
                    #Check that the closest MJDOBS is closer than 2 hours
                    if abs(MJDOBS_closest-MJDOBS) > 2./24.:
                        # Skip if we couldn't find a matching date.
                        return [],[]

                    for obj_id in np.where(MJDOBS_arr == MJDOBS_closest)[0]:
                        try:
                            pa = float(GOI_list[obj_id,pa_id])
                            radius = float(GOI_list[obj_id,sep_id])
                            if IWA is not None:
                                if 1./pix2as*radius < IWA:
                                    continue
                            if OWA is not None:
                                if 1./pix2as*radius > OWA:
                                    continue
                            status = str(GOI_list[obj_id,STATUS_id])
                            # print(status, include_speckles or (status in ["Planet","Background","Candidate","Unknown","Brown Dwarf"]))
                            if include_speckles or (status in ["Planet","Background","Candidate","Unknown","Brown Dwarf"]):
                                pa_vec.append(pa)
                                sep_vec.append(radius)
                                x_max_pos = float(1./pix2as*radius)*np.cos(np.radians(90+pa))
                                y_max_pos = float(1./pix2as*radius)*np.sin(np.radians(90+pa))
                                x_vec.append(x_max_pos)
                                y_vec.append(y_max_pos)
                                row_vec.append(y_max_pos+center[1])
                                col_vec.append(x_max_pos+center[0])
                        except:
                            print("Missing data in GOI database for {0}".format(_object_name))

    if not ignore_fakes:
        for fake_id in range(100):
            try:
                pa = fakeinfohdr["FKPA{0:02d}".format(fake_id)]
                radius = pix2as*fakeinfohdr["FKSEP{0:02d}".format(fake_id)]
                if IWA is not None:
                    if 1./pix2as*radius < IWA:
                        continue
                if OWA is not None:
                    if 1./pix2as*radius > OWA:
                        continue
                pa_vec.append(pa)
                sep_vec.append(radius)
                x_max_pos = float(1./pix2as*radius)*np.cos(np.radians(90+pa))
                y_max_pos = float(1./pix2as*radius)*np.sin(np.radians(90+pa))
                x_vec.append(x_max_pos)
                y_vec.append(y_max_pos)
                row_vec.append(y_max_pos+center[1])
                col_vec.append(x_max_pos+center[0])
            except:
                continue


    if pa_sep:
        return sep_vec,pa_vec
    elif xy:
        return x_vec,y_vec
    else:
        return row_vec,col_vec


def make_GOI_list(outputDir,GOI_list_csv,GPI_TID_csv):
    """
    Generate the GOI files from the GOI table and the TID table (queried from the database).

    outputDir: Output directory in which to save the GOI files.
    GOI_list_csv: Table with the list of GOIs (including separation, PA...). delimiter=','
    GPI_TID_csv: Table giving the TID code for a given object name. delimiter=','
    :return: One .csv file per target for which at list one GOI exists. delimiter=';'
            The filename follows: [object]_GOI.csv. For e.g. c_Eri_GOI.csv.
    """
    with open(GOI_list_csv, 'r') as csvfile_GOI_list:
        GOI_list_reader = csv.reader(csvfile_GOI_list, delimiter=',')
        GOI_csv_as_list = list(GOI_list_reader)
        attrib_name = GOI_csv_as_list[0]
        GOI_list = np.array(GOI_csv_as_list[1:len(GOI_csv_as_list)])

        TID_id = attrib_name.index("TID")
        TID_GOI = GOI_list[:,TID_id]
        TID_unique = np.unique(TID_GOI)

        with open(GPI_TID_csv, 'r') as csvfile_TID:
            TID_reader = csv.reader(csvfile_TID, delimiter=',')
            TID_csv_as_list = list(TID_reader)
            TID_csv_as_nparr = np.array(TID_csv_as_list)[1:len(TID_csv_as_list),:]
            TID_campaign = np.ndarray.tolist(TID_csv_as_nparr[:,0])
            star_campaign = np.ndarray.tolist(TID_csv_as_nparr[:,1])
            #print(TID_campaign)
            #print(star_campaign)

            dict_matching_TID_to_name = {}
            for TID_it in TID_unique:
                star_raw_name = star_campaign[TID_campaign.index(TID_it)]
                star_name = star_raw_name.replace(" ","_")
                dict_matching_TID_to_name[TID_it] = star_name
            #print(dict_matching_TID_to_name)

        for TID_it in TID_unique:
            where_same_star = np.where(TID_GOI == TID_it)

            with open(outputDir+os.path.sep+dict_matching_TID_to_name[TID_it]+'_GOI.csv', 'w+') as csvfile:
                csvwriter = csv.writer(csvfile, delimiter=';')
                table_to_csv = [attrib_name]+np.ndarray.tolist(GOI_list[where_same_star[0],:])#.insert(0,attrib_name)
                #print(table_to_csv)
                csvwriter.writerows(table_to_csv)